/************************************************************************************
 * masterTasks_sct.c
 *   
 *  synopsis: This file contains tasks which run on the MDSP, and are optimized
 *            for SCT use. Typically these coordinate and monitor the SDSPs as
 *            they process data.
 *
 *  in this file: histoCtrlInit, histoCtrlNewBin, histoCtrlWaitExp, histoCtrlWaiting,
 *                histoCtrlPrep, histoCtrlDone, histoCtrlTask.
 *
 *  related files:
 *   masterTasks.c:  MDSP tasks common to both SCT & Pixels.
 *   taskManager.c:  Contains routines for starting, stopping & managing tasks.
 *   slaveTasks.c:   Contains the slave DSP tasks.
 *   histogram.h:    Contains the definitions for the histogram control structures
 *                   on the master & slave DSPs.
 ************************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>

#include <csl.h>
#include "resources.h"
#include "comRegDfns.h"
#include "registerIndices.h"
#include "accessSlave.h"
#include "histogram.h"
#include "accessRegister.h"

extern char genStr[];
extern far unsigned char heap[];
extern TIMER_Handle timer1;  /* general purpose timer, started in main() */

extern far struct TaskMgrCtrl taskMgrCtrl;
extern far RodModeCfg rodModeCfg[];
extern far Module moduleConfigSet[N_MODULE_CONFIG_SETS][N_TOTMODULES];
extern far CmdBuff cmdBuffer[];
extern far CfgTiming cfgTiming;
extern far MaskConfigData maskConfigData[];

#pragma CODE_SECTION(addDspInfo,         "icode");
#pragma CODE_SECTION(getDspStatus,       "icode");
//#pragma CODE_SECTION(updateEventStats,   "icode");
#pragma CODE_SECTION(initSerialStreams,  "xcode");
//#pragma CODE_SECTION(setPending,         "xcode");
#pragma CODE_SECTION(scanWait,           "xcode");
#pragma CODE_SECTION(getTaskCtrlBits,    "xcode");
#pragma CODE_SECTION(resetModules,       "xcode");
#pragma CODE_SECTION(buildScanTrigger,   "icode");
#pragma CODE_SECTION(modifyTriggerDelay, "icode");

#pragma CODE_SECTION(newTrapSeries,        "icode");
#pragma CODE_SECTION(scanAttnHandle,       "icode");
#pragma CODE_SECTION(scanEventCheck,       "icode");
#pragma CODE_SECTION(scanErrHandle,        "icode");
#pragma CODE_SECTION(updateScanStatistics, "icode");
#pragma CODE_SECTION(scanExtrapolate,      "icode");

#pragma CODE_SECTION(histoCtrlInit,      "xcode");
#pragma CODE_SECTION(histoCtrlSdspSetup, "icode");
#pragma CODE_SECTION(histoCtrlNewBin,    "icode");
#pragma CODE_SECTION(histoCtrlWaitExp,   "icode");
#pragma CODE_SECTION(histoCtrlWaiting,   "icode");
#pragma CODE_SECTION(histoCtrlPrep,      "icode");
#pragma CODE_SECTION(histoCtrlDone,      "icode");

//#pragma CODE_SECTION(histoCtrlWaitExp,   "icode");
#pragma CODE_SECTION(histoCtrlTask,      "icode");

#pragma CODE_SECTION(scanTaskOp,         "xcode");
#pragma CODE_SECTION(scanTaskStates,     "xcode");

#pragma DATA_SECTION(hc,         "idata");
#pragma DATA_SECTION(scanStatus, "xpdata");
#pragma DATA_SECTION(varRange,   "xpdata");
MasterHistoCtrl hc;
ScanStatus scanStatus;
MDAT32 varRange[2][HISTOGRAM_BIN_MAX];

//Structures needed for SDSP initialization:
#pragma DATA_SECTION(hc_ets_i,      "xpdata");
#pragma DATA_SECTION(hc_ets_o,      "xpdata");
#pragma DATA_SECTION(hc_hsetup_i,   "xpdata");
#pragma DATA_SECTION(hc_hsetup_o,   "xpdata");
#pragma DATA_SECTION(hc_htask_i,    "xpdata");
EventTrapSetupIn   hc_ets_i;
EventTrapSetupOut  hc_ets_o;
HistogramSetupIn   hc_hsetup_i;
HistogramSetupOut  hc_hsetup_o;
StartTaskIn        hc_htask_i;

//dpsf: make this an ETS sub-structure:
RouterTrapParams trapParams[2];

/* file-level function declarations */
void  resetModules(UINT8 sdsp, UINT8 ecr, UINT8 bcr, UINT8 send, UINT8 makeTrig);
void  buildScanTrigger(UINT8 init);
void  modifyTriggerDelay(UINT8 scanParam);

INT32 startHistoSlaves(UINT8 slvBits);
void  addDspInfo(UINT8 buffer, UINT8 sdsp);
INT32 getDspStatus(UINT8 slv, UINT8 nTrigPending);
void  updateEventStats(UINT8 slv);
INT32 initSerialStreams(void);
void  getTaskCtrlBits(void);
void  scanWait(UINT32 messageId, UINT32 sdsp);
void  setPending(UINT8 sdsp);

void  newTrapSeries(void);
INT32 scanEventCheck(UINT8 wait);
INT32 scanErrHandle(void);
void  updateScanStatistics(void);
void  scanExtrapolate(void);

//Scan task states:
INT32 histoCtrlInit(TaskInput *tsi);
INT32 histoCtrlSdspSetup(TaskInput *tsi);
INT32 histoCtrlNewBin(void);
INT32 histoCtrlWaitExp(void);
/* event loop is time sensitive & done in the histoCtrlTask */
INT32 histoCtrlWaiting(void);
UINT8 histoCtrlPrep(void);
INT32 histoCtrlDone(void);

INT32 scanTaskOp(union TASK_STRUCTURES_OUT *tso, UINT32 opFlags, UINT8 *state);
INT32 scanTaskStates(union TASK_STRUCTURES_IN *tsi, UINT8 *state);

/* Histogram Control Task state-machine states: */
/* TASK_INIT == 1 */
#define  HCTRL_SDSP_SETUP  2
#define  HCTRL_NEWBIN      3
#define  HCTRL_WAITEXP     4
#define  HCTRL_TRIGGERS    5
#define  HCTRL_WAITING     6
#define  HCTRL_PREP        7

#define HCTRL_TIMEOUT        0x00990000   //~1 sec.
#define MAX_RECOVER_ATTEMPTS 10

UINT32 efbEvtCnt[N_MODULE_GROUPS]= {0,0,0,0,  0,0,0,0};

/* Three extra (internally used) configuration parameters which define actions
   which the MDSP scanning routine will take: */
#define CONFIG_BOC           29
#define CONFIG_NONE          30
#define CONFIG_ROD_TRIGGER   31

//dpsf patch: re-do SP code.
CmdList nullCmdList, *trigPtr[2];
/* Message IDs for scanWait: */
enum {
	SCAN_INIT_CFG= 0,  SCAN_END_INIT,

	SCAN_NB_INIT= 10, SCAN_SDSP_SETUP, SCAN_STAGE_CFG, SCAN_STAGE_SEND, SCAN_CFG_VARIABLE,
	SCAN_SEND_CFG, SCAN_NB_ECR_BCR, SCAN_WAITEXP, SCAN_WAITEXP_GOOD,

	SCAN_DATA_PREP_INIT= 50,  SCAN_DATA_PREP_ADD, SCAN_DATA_PREP_SEND,
	SCAN_RESET_SETMASKS= 100, SCAN_RESET_SEND, SCAN_RESET_EXIT,
	SCAN_TRIGGER_ENTRY=  150, SCAN_SDSP_READY_CHECK, SCAN_SDSP_READY, SCAN_TRIGGER,

	SCAN_EVENT_CHECK_START= 200, SCAN_REXMIT_MASKS_SET, SCAN_ERROR_RECOVERY,
	SCAN_AUTO_STALL, SCAN_EVENT_CHECK, SCAN_EVENT_CORRECT
};

/************************************************************************************
 * getTaskCtrlBits:  Retrieve run-time task control bits.
 ************************************************************************************/
void getTaskCtrlBits(void) {
	/* Get expert run-time flags. */
	hc.manual=      hc.soft_bc_reset=  hc.autoStall=
	hc.trigPause=  hc.binPause=       hc.stagePause=  FALSE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_STEP_CTRL))     hc.manual= TRUE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_SOFT_BC_RESET)) hc.soft_bc_reset= TRUE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_AUTO_STALL))    hc.autoStall= TRUE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_TRIG_PAUSE))    hc.trigPause= TRUE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_BIN_PAUSE))     hc.binPause= TRUE;
	if (GET_RBIT(DIAGNOSTIC_REG, DR_STAGE_PAUSE))   hc.stagePause= TRUE;
}

/************************************************************************************
 * scanWait:  scanWait prints out a message and waits for the operator to set a bit
 *            in the diagnostic register before continuing. Note that during the
 *    waitRegister routine input from primitive lists is possible so long as the
 *    primitive list containing the start-task primitive has finished executing;
 *    generally this means any state except the INIT state.
 ************************************************************************************/
void scanWait(UINT32 messageId, UINT32 sdsp) {

	switch(messageId) {

		case SCAN_INIT_CFG:
			sprintf(genStr, "INIT: Pause after module reconfig.\n");
			break;

		case SCAN_END_INIT:
			//histoReport();
			sprintf(genStr, "Pause before leaving INIT.\n");
			break;


		case SCAN_NB_INIT:
			sprintf(genStr, "NEWBIN: Pausing.\n");
			break;

		case SCAN_SDSP_SETUP:
			//histoReport();
			sprintf(genStr, "NEWBIN: about to set SDSP histogramming control registers.\n");
			break;

		case SCAN_STAGE_CFG:
			sprintf(genStr, "NEWBIN: setting new mask stage in modules.\n");
			break;

		case SCAN_STAGE_SEND:
			sprintf(genStr, "NEWBIN: sending new mask stage to modules.\n");
			break;

		case SCAN_CFG_VARIABLE:
			sprintf(genStr, "NEWBIN: load scan variable.\n");
			break;

		case SCAN_SEND_CFG:
			sprintf(genStr, "NEWBIN: sending module configuration data.\n");
			break;

		case SCAN_WAITEXP:
			sprintf(genStr, "WAITEXP: dammit.\n");
			break;

		case SCAN_WAITEXP_GOOD:
			sprintf(genStr, "WAITEXP: leaving.\n");
			break;


		case SCAN_RESET_SETMASKS:
			sprintf(genStr, "resetModules: setting masks next.\n");
			break;

		case SCAN_RESET_SEND:
			sprintf(genStr, "resetModules: sending reset command.\n");
			break;

		case SCAN_RESET_EXIT:
			sprintf(genStr, "resetModules: exit.\n");
			break;


		case SCAN_NB_ECR_BCR:
			sprintf(genStr, "NEWBIN: about to send soft & BC reset.\n");
			break;


		case SCAN_TRIGGER_ENTRY:
			sprintf(genStr, "PULSING: paused.\n");
			break;

		case SCAN_SDSP_READY_CHECK:
			//histoDspReport(sdsp);   if (hc.dspInfo[sdsp].paired) histoDspReport(hc.dspInfo[sdsp].pair);
			sprintf(genStr, "PULSING: checking SDSP #%d condition.\n", sdsp);
			break;

		case SCAN_SDSP_READY:
			sprintf(genStr, "PULSING: SDSP #%d ready for trigger, set Rod Registers.\n", sdsp);
			break;

		case SCAN_TRIGGER:
			sprintf(genStr, "PULSING for SDSP #%d.\n", sdsp);
			break;


		case SCAN_EVENT_CHECK_START:
			//histoDspReport(sdsp);   if (hc.dspInfo[sdsp].paired) histoDspReport(hc.dspInfo[sdsp].pair);
			sprintf(genStr, "PULSING: checking event(s) on SDSP #%d\n", sdsp);
			break;

		case SCAN_REXMIT_MASKS_SET:
			sprintf(genStr, "PULSING: set masks for SDSP %d, will send ECR-BCR next.\n", sdsp);
			break;

		case SCAN_ERROR_RECOVERY:
			sprintf(genStr, "PULSING: sent ECR-BCR to SDSP #%d, will send CAL-L1A next.\n", sdsp);
			break;

		case SCAN_AUTO_STALL:
			//histoGroupReport(sdsp);   histoDspReport(sdsp);
			sprintf(genStr, "PULSING: Pause for error event inspection on SDSP #%d\n", sdsp);
			break;

		case SCAN_EVENT_CHECK:
			//histoGroupReport(sdsp);   histoDspReport(sdsp);
			sprintf(genStr, "PULSING: pause after checking event on SDSP #%d.\n", sdsp);
			break;

		case SCAN_EVENT_CORRECT:
			sprintf(genStr, "PULSING: pause before correcting missing trailer on SDSP #%d.\n", sdsp);
			break;

		default:
			sprintf(genStr, "Pause.\n");
		break;
	}

	newInformation(__FILE__, __LINE__, genStr);
	waitRegister(DIAGNOSTIC_REG, DR_STEP_TRIGGER, CONTEXT_TASK);
}


/************************************************************************************
 * resetModules:  Reset (all) the modules by sending a soft-BC reset combination to
 *                them. Afterward, re-build the trigger stream(s).
 ************************************************************************************/
void resetModules(UINT8 sdsp, UINT8 ecr, UINT8 bcr, UINT8 send, UINT8 makeTrig) {
	INT32 errorCode, j, k;

	if (send) {
		if (hc.manual) scanWait(SCAN_RESET_SETMASKS, sdsp);
		//switchLinkMasks(((DATA_LINK)+(COMMAND_LINK)), MASK_SET_INIT);
		setLinkMasks(0, SP_BOTH, COMMAND_LINK, INIT_MASK, MASK_SET_WORK1);

		for (j= 0; j < N_MODULE_GROUPS; ++j) {
			for (k=0; k<N_TOTMODULES; ++k) {
			if ((hc.moduleCfg[k].present) && (hc.moduleCfg[k].groupId == j)) {
				setLinkMasks(k, hc.groupInfo[j].serialPort, COMMAND_LINK_ON,
				             UPDATE_MASK, MASK_SET_WORK1);
			}}
		}

		setLinkMasks(0, SP_BOTH, COMMAND_LINK, SET_MASK, MASK_SET_WORK1);
	}

	hc.postConfigCmdList.cmd[0]=  (ecr)? SOFT_RESET:NO_CMD;
	hc.postConfigCmdList.cmd[1]=  DELAY;       hc.postConfigCmdList.data[1]= 0x10;
	hc.postConfigCmdList.cmd[2]=  (bcr)? BC_RESET:NO_CMD;
	errorCode= makeStreamSet(hc.cmdListPtr[0], hc.cmdListPtr[1],
	                         TRUE, ALL_CHIPS, PRIMARY_FIBRE, 0, NULL,
	                         &hc.ptr, &hc.nSets, send);

	if (hc.manual & send) scanWait(SCAN_RESET_SEND, sdsp);
	if (send) serialStreamOut(0, (TransData *) hc.ptr, hc.nSets);  

	/* Typically, this routine is called between bins & sends an ECR and/or BCR if the
	   user has indidicated that this is desired. To keep it simple the combination is
	   still sent if and ECR is sent. Note that when an ECR is sent out to any module, it
	   resets the Bunch Crossing counter (on both ROD & module) as well. Occasionally a
	   module can send a bad event to one or more of the SDSPs during a scan, in which
	   case the routine is called from the error recovery code. In that case, the other
	   SDSPs (which may have good events) must not update their counters since their
	   modules will receive an ECR/BCR combination as well; this is the purpose of the
	   ecrSent flag. It is reset at the beginning of every event-checking loop. */
	for (j=0; j<N_SDSP; ++j) {hc.dspInfo[j].L1id[0]= hc.dspInfo[j].L1id[1]= 0; }
	hc.ecrSent= TRUE;

	/* Create the trigger stream. Note that if send is false, the intent was to
	   add the ecr/bcr combination as the initial part of the trigger; therefore,
	   no initialization is done on the command buffer. */
	if (makeTrig) {
		errorCode= makeStreamSet(&hc.triggerSequence[0], &hc.triggerSequence[1],
		                         send, ALL_CHIPS, PRIMARY_FIBRE, 0, NULL,
		                         &hc.ptr, &hc.nSets, TRUE);
	}

	if (send) {
//dpsf: patch.
		if (!GET_RBIT(DIAGNOSTIC_REG, DR_USE_MDSP_MASK_LUT))
			writeRegisterDirect(FE_MASK_LUT_SELECT, 3, 0, sdsp);
		else {
			setLinkMasks(0, SP_BOTH, COMMAND_LINK, SET_MASK, MASK_SET_INIT);
			switchLinkMasks(((DATA_LINK)+(COMMAND_LINK)), sdsp);
		}
	}
	if (hc.manual) scanWait(SCAN_RESET_EXIT, sdsp);
}

/************************************************************************************
 * buildScanTrigger:  Builds the calibration-pulse/delay/L1A serial streams sent
 *                    during the pulsing loop.
 ************************************************************************************/
void buildScanTrigger(UINT8 init) {
	UINT8 i, j, spi, spf;
	
	if (init) { //Routine called during initialization:
		for (i=0; i<2; ++i) {
			for (j=0; j<N_CMD_LIST_CMDS; ++j) {
				hc.triggerSequence[i].cmd[j]= NO_CMD;	
				hc.triggerSequence[i].data[j]= 0;	
			}
		}
	}

	//Newbin called this routine:
	if (hc.scan->serial.port == SP_BOTH) {spi= 0;                    spf= SP_BOTH; }
	else                                 {spi= hc.scan->serial.port; spf= spi +1;  }

	for (i=spi; i<spf; ++i) {
		/* If both serial ports are being used, the port 1 trigger is issued
		   10 cycles (abribrarily; separation must be at least 6 cycles) after
		   the port 0 trigger */
		if (hc.scan->serial.port == SP_BOTH) {
			hc.triggerSequence[i].cmd[0]= DELAY;
			hc.triggerSequence[i].data[0]= i*10;
		}
		hc.triggerSequence[i].cmd[1]=  CALIBRATION_PULSE;	
		hc.triggerSequence[i].cmd[2]=  DELAY;		
		hc.triggerSequence[i].data[2]= hc.status->currentDelay;
		hc.triggerSequence[i].cmd[3]=  L1_TRIGGER;
	}
}

/************************************************************************************
 * modifyTriggerDelay:  Finds & modifies the trigger delay(s) inside the user-defined
 *                      trigger sequences.
 ************************************************************************************/
void modifyTriggerDelay(UINT8 scanParam) {
	UINT8 sp, i, cnt, spi, spf;
	
	if (hc.scan->serial.port == SP_BOTH) {spi= 0;                    spf= SP_BOTH; }
	else                                 {spi= hc.scan->serial.port; spf= spi +1;  }

	for (sp=spi; sp<spf; ++sp) {
		for (i= cnt= 0; i<N_CMD_LIST_CMDS; ++i) {
			if (hc.triggerSequence[sp].cmd[i] == DELAY) {
				++cnt;
				if (  ((scanParam == ST_TRIG_DELAY1) && (cnt == 1))
				    ||((scanParam == ST_TRIG_DELAY2) && (cnt == 2)) )
					hc.triggerSequence[sp].data[i]= hc.status->currentDelay;
			}
		}
	}
}



#define DSP_BUSY -1
#define DSP_READY 1

/************************************************************************************
 * addDspInfo
 *   
 *   synopsis: Place information about current status of the MDSP histogramming
 *             control and of the SDSP trapping into one of the text buffers.
 ************************************************************************************/
void addDspInfo(UINT8 buffer, UINT8 sdsp) {
	UINT8 idSdsp, port;
	
	if (hc.dspInfo[sdsp].active) idSdsp= sdsp;
	else                         idSdsp= hc.dspInfo[sdsp].pair;

	if (hc.dspInfo[idSdsp].ports == 2) port= SP1;
	else                               port= SP0;

	sprintf(genStr, "\n %s%d%s%d%s%d%s%d%s%d%s %s0x%08x%s %s0x%08x%s %s0x%08x%s %s0x%08x%s", 
		"   MDSP:  bin #", hc.status->currentBin[0],
		"   SDSP #", sdsp,
		"   evt. #", hc.dspInfo[idSdsp].nEvents,
		"   err. #", hc.dspInfo[idSdsp].nErrors,
		"  (L1-ID ", hc.dspInfo[idSdsp].L1id[port], ").\n\n",

		"   TCmd= ", hc.dspInfo[sdsp].trapCmd, ",\n",
		"   Current TS0= ", getSlvReg(sdsp, TRAPSTAT_REG1_0), ".\n",
		"   Current TS1= ", getSlvReg(sdsp, TRAPSTAT_REG1_1), ".\n",
		"   Current TS2= ", getSlvReg(sdsp, TRAPSTAT_REG1_2), ".\n");

	if (buffer == ERR_BUFF) addErrorInfo(__FILE__, __LINE__, genStr);
	else                    addInformation(__FILE__, __LINE__, genStr);
}

/************************************************************************************
 * getDspStatus
 *   
 *   synopsis: Retrieves the number of frames currently waiting to be proccessed
 *             inside a slave DSP via pointers, and returns a status indicating
 *     whether or not a new event (or events) would be likely to force the DSP to
 *     use slow memory. Other variables inside the hc (histogram control) structure
 *     are set as well, for later reference.
 ************************************************************************************/
INT32 getDspStatus(UINT8 sdsp, UINT8 nTriggers) {
	INT32 status;
	static UINT32 cnt= 0;
	register UINT32 a, b, c, d;
	register UINT32 x, y, z;

//dpsf: If the appropriate bit is set in the DR, check to make sure proc flag
//dpsf: is down before allowing ready. If not, increment RR3:
	if (GET_RBIT(DIAGNOSTIC_REG, DR_PROCWAIT)) {
		x= getSlvRegBit(sdsp, HSTATUS_REG_0, HSR0_SLV_PROC);
		if (x) return DSP_BUSY;
	}

	x= getSlvReg(sdsp, TRAPSTAT_REG1_2);
	z= hc.dspInfo[sdsp].avgEvtLen>>8;

	hc.dspInfo[sdsp].iTail= a= GET_VFLD(x, TSR2_IFRAME_TAIL, TSR2_IFRAME_TAIL_W);
	hc.dspInfo[sdsp].xTail= b= GET_VFLD(x, TSR2_XFRAME_TAIL, TSR2_XFRAME_TAIL_W);
	hc.dspInfo[sdsp].iHead= c= GET_VFLD(x, TSR2_IFRAME_HEAD, TSR2_IFRAME_HEAD_W);
	hc.dspInfo[sdsp].xHead= d= GET_VFLD(x, TSR2_XFRAME_HEAD, TSR2_XFRAME_HEAD_W);

	if (a < c) a+= 32;                   if (b < d) b+= 224;
	hc.dspInfo[sdsp].nIdram= a= a -c;    hc.dspInfo[sdsp].nSdram= b= b -d;
	hc.dspInfo[sdsp].trapStat[2]= x;

	status= ( (a == 0) || ((a +nTriggers*z)<28) )? DSP_READY:DSP_BUSY;
	//if (!status) greenLed_toggle;

	if (status) {
		WRITE_REG(RESERVED_REG_2, a);
		WRITE_REG(RESERVED_REG_1, cnt);
		cnt= 0;
	}
	else {
		if (cnt == 0) WRITE_REG(RESERVED_REG_3, a);
		++cnt;
	}

	return status;
}

#if 0
/************************************************************************************
 * setPending
 *   
 *   synopsis: Sets the pending bit in the trap command register on the selected
 *             SDSP (and its paired twin, if any). The pending bit then resets the
 *             event counters on the SDSP(s) upon arrival of the next event.
 ************************************************************************************/
void setPending(UINT8 sdsp) {
	UINT8 twin;

	twin= hc.dspInfo[sdsp].pair;
	/* shouldn't need doing-- check & make more efficient */
	//while (getSlvRegBit(slv, TRAP_CMD_STAT, TCSR_ISR_ACTIVE));
	setSlvReg(sdsp, TRAP_CMD_STAT,
	          (  (hc.dspInfo[sdsp].trapCmdStat | (1<<TCSR_ISR_PENDING))
	           & ( ~((1<<TCSR_ISR_ACTIVE)+(1<<TCSR_HEADER)) )          )  );

	if (hc.dspInfo[sdsp].paired)
		setSlvReg(twin, TRAP_CMD_STAT,
		          (  (hc.dspInfo[twin].trapCmdStat|(1<<TCSR_ISR_PENDING))
		           & ( ~((1<<TCSR_ISR_ACTIVE)+(1<<TCSR_HEADER)) )        ) );
}
#endif


#if 0
/************************************************************************************
 * updateEventStats
 *   synopsis: Updates the MDSP's displayed event statistics.
 ************************************************************************************/
void updateEventStats(UINT8 sdsp) {
	register UINT32 a, w, x, y, z;

	if      (sdsp == 0) a= HSSTAT_REG_0;
	else if (sdsp == 1) a= HSSTAT_REG_1;
	else if (sdsp == 2) a= HSSTAT_REG_2;
	else if (sdsp == 3) a= HSSTAT_REG_3;

	w= 	hc.dspInfo[sdsp].avgEvtLen;
	x= getSlvReg(sdsp, TRAPSTAT_REG_0);
	y= ++hc.dspInfo[sdsp].nGoodEvt;
	z= GET_VFLD(x, TSR0_EVT_WORD_CNT, TSR0_EVT_WORD_CNT_W);

	x= ((y-1)*w +z)/y;

	hc.evtLen= (UINT16) z;
	hc.dspInfo[sdsp].avgEvtLen= x;

	WRITE_REG(a, y);
}
#endif

/************************************************************************************
 * newTrapSeries
 *   
 *   synopsis: Creates a new trap series on all SDSPs. The trap series controls how
 *             the counters on the SDSPs are incremented when they receive a new
 *      event header. If it is the same as their current trap series number they
 *      will simply add to the event count when the event has ended. If the trap
 *      series differs from their current trap series, they will zero out all of
 *      the event counters and set a new trap series number to match the input. For
 *      simplicity, all the SDSPs are given the same trap series number.
 ************************************************************************************/
void newTrapSeries(void) {
	register UINT32 sdsp, x, y, i;
	UINT32 series= 0xffff;
	
	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (!(hc.slvBits & (1<<sdsp))) continue;
		x= getSlvReg(sdsp, TRAPSTAT_REG1_0);
		//Remove this SDSP's series # from the possible series bitfield: 
		y= GET_VFLD(x, TSR0_EVTSERIES, TSR0_EVTSERIES_W);
		series&= ~(1<<y);
	}

	/* The new series is the lowest possible series that is higher than the previous
	   value (if any). It wraps at 16 as 15 is the maximum possible series value. */
	do {
		++hc.trapSeries;
		hc.trapSeries&= 0xf;
	}
	while (!(series & (1<<hc.trapSeries)));

	//Now assign the new series # to the SDSPs:
	x= hc.trapSeries<<TCMD_EVTSERIES;  //It is the only field so we can just write it.
	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (!(hc.slvBits & (1<<sdsp))) continue;
		setSlvReg(sdsp, TRAP_CMD_REG1, x);
		for (i=0; i<2; ++i) {
			hc.dspInfo[sdsp].seriesEvents[i]= 0;
			hc.dspInfo[sdsp].seriesErrors[i]= 0;
		}
	}

}

/************************************************************************************
 * scanAttnHandle
 *   
 *   synopsis: Analyzes and corrects the situation if any SDSP has its attention
 *             flag raised. This can be caused by either an error event arriving
 *      on the SDSP (in which case we simply lower the flag as the error will be
 *      handled once seen), or a SDSP is receiving data faster than it can process
 *      it. In the latter case we adjust the MDSP's speed throttle.
 ************************************************************************************/
 INT32 scanAttnHandle() {
	INT32 returnCode= SUCCESS, status;
	register UINT32 sdsp;
	register UINT32 dt, t0, x;
	register UINT8  a, b, c, d, err, errTot= 0;
	register int procTime;

 	/* The raw flag is inverted from the usual meaning (i.e. 0xf => no issues) so
 	   first we invert it. The SDSP aux. status reg. (2) stores the attention flags. */
	hc.sdspAttention= (~hc.sdspAttention) & 0xf;
	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (!(hc.sdspAttention & (1<<sdsp))) continue;
		x= getSlvReg(sdsp, STATUS_REG_2);
		a= x & (1<<ATTN_IFRAME_CNT);   b= x & (1<<ATTN_SLOW_DATA);
		c= x & (1<<ATTN_ERR_EVT);      d= x & (1<<ATTN_BUFF_OVF);

		if (a) { //Regulate trigger:
			//Introduce the approx. delay into the trigger loop:
			procTime= hc.dspInfo[sdsp].avgProcTime;
			if (procTime == 0) procTime= getSlvReg(sdsp, SDSP_HSTATUS_REG_3);
			hc.trigDelay= (procTime > hc.trigDelay)? procTime:hc.trigDelay;
			
			//Wait until the SDSP internal frame count dips below 25% :
			t0= TIMER_getCount(timer1);
			do {
				status= getDspStatus(sdsp, 0);
				dt= delta_t(t0);
			} while ( (hc.dspInfo[sdsp].nIdram > 8) && (dt < HCTRL_TIMEOUT) );

			if (dt >= HCTRL_TIMEOUT) { //Issues on SDSP:
				sprintf(genStr, "%s%d%s", "SDSP #",sdsp,
				       ": timed out while waiting for event backlog to clear.\n");
				newError(&returnCode, EVENT_ERR, FATAL_ERR, "scanErrHandle", genStr,
				         __FILE__, __LINE__);
	
				addDspInfo(ERR_BUFF, sdsp);
				return returnCode;
			}
		}

		//Clear the attention flag on the SDSP:
		x= getSlvReg(sdsp, COMMAND_REG_0);
		x|=  (0xf<<16);
		setSlvReg(sdsp, COMMAND_REG_0, x);

		writeRegister(INTRPT_TO_SLV, 4, 0, (1<<sdsp));
		delay(10);
		writeRegister(INTRPT_TO_SLV, 4, 0, 0);

		x&= ~(0xf<<16);
		setSlvReg(sdsp, COMMAND_REG_0, x);
	}

	return returnCode;
}


/************************************************************************************
 * scanEventCheck
 *   
 *   synopsis: Checks that the events which have arrived on the SDSPs so far contain
 *             no errors, by checking the trapping status registers on all the SDSPs
 *      participating in the scan. If there are any errors, it will simply return
 *      the number of errors and let other routines decide what to do about them.
 ************************************************************************************/
INT32 scanEventCheck(UINT8 wait) {
	INT32 returnCode= SUCCESS, status;
	register UINT32 sdsp;
	register UINT32 dt, t0, x;
	register UINT8  a, b, c, err, errTot= 0;
	
	t0= TIMER_getCount(timer1);

	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (!(hc.pulsing & (1<<sdsp))) continue;

		if (wait) { //Wait while last event is still transmitting, with a timeout.
			do {
				dt= delta_t(t0);
				x= getSlvReg(sdsp, TRAPSTAT_REG1_0);
				a=  GET_VBIT(x,TSR0_HEADER);
				b=  (GET_VBIT(x,TSR0_ISR_ACTIVE) || GET_VBIT(x,TSR0_TRANSMIT));
				c=  GET_VBIT(x,TSR0_TRAILER);
			} while ( ((!a) || (b)) && (!c) && (dt < HCTRL_TIMEOUT) );
		}

		hc.dspInfo[sdsp].trapStat[0]= getSlvReg(sdsp, TRAPSTAT_REG1_0);
		hc.dspInfo[sdsp].trapStat[1]= getSlvReg(sdsp, TRAPSTAT_REG1_1);
		hc.dspInfo[sdsp].trapStat[2]= getSlvReg(sdsp, TRAPSTAT_REG1_2);

		//Get the # of events/errors seen in this trap series:
		hc.dspInfo[sdsp].seriesEvents[1]=
			GET_VFLD(hc.dspInfo[sdsp].trapStat[1], TSR1_EVT_COUNT,TSR1_EVT_COUNT_W);

		hc.dspInfo[sdsp].seriesErrors[1]=
			GET_VFLD(hc.dspInfo[sdsp].trapStat[1], TSR1_ERR_COUNT,TSR1_ERR_COUNT_W);
			
		//Add any errors to the error count:
		err= (hc.dspInfo[sdsp].seriesErrors[1] -hc.dspInfo[sdsp].seriesErrors[0]);
		hc.dspInfo[sdsp].delErr= err;
		errTot+= err;
	}

	return errTot; //Return the number of errors.
}

/************************************************************************************
 * scanErrHandle
 *   
 *   synopsis: Handle any errors which occur during a scan. If the routine manages
 *             to correct the error, it will return SUCCESS, otherwise it will
 *      exit with an error code and the scan will abort. The action taken depends
 *      upon the detector & error(s). Many times it will re-send the trigger, after
 *      1st sending a soft-reset command to all modules. Events are re-transmitted
 *      up to 10 times if there are errors, after which the routine assumes that
 *      there is an incurable problem with the electronics and exits.
 ************************************************************************************/
INT32 scanErrHandle(void) {
	INT32 returnCode= SUCCESS, status;
	UINT32 t0, dt, attempts;
	UINT8 stillErrors, failed= FALSE;
	UINT32 err[4], nErrs;
	register UINT32 icnt1, icnt2, cnt, ready, sdsp, twin;
	
	//Wait for the data path to clear:
	t0= TIMER_getCount(timer1);
	status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt1= dt;
	do {
		delay(20); //2 us delay.
		status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt2= dt;
		ready= (icnt2 == icnt1); 
		icnt1= icnt2;
	} while ((!ready) && ((dt= delta_t(t0)) < HCTRL_TIMEOUT));
	cnt= delta_t(hc.trigTime);  icnt2= cnt>>4;
	delay(icnt2);

	/* Now wait for the events to arrive on the SDSPs (re-call the event-checking
	   routine, this time requesting that it wait until the events have fully
	   arrived. When this routine is called from the trigger loop, normally it
	   does not bother waiting. This is done here to avoid disturbing the trigger
	   loop for abnormal conditions. */
	scanEventCheck(TRUE);

	//1st update the scan's statistics & counters with any events/errors:
	updateScanStatistics();

	do {
		for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
			if (!(hc.pulsing & (1<<sdsp))) continue;

			err[sdsp]= hc.dspInfo[sdsp].delErr;
			if (!err[sdsp]) continue;
			
			sprintf(genStr, "%s%d%s", "SDSP #",sdsp,": error events- correcting.\n\n");
			newError(&returnCode, EVENT_ERR, ERROR_0, "scanErrHandle", genStr,
			         __FILE__, __LINE__);

			addDspInfo(ERR_BUFF, sdsp);
		}

		//Issue a soft reset to all the modules, and re-create the trigger stream:
		resetModules(0, TRUE, TRUE, TRUE, TRUE);	
		
		newTrapSeries();   //Give the SDSPs a new trap series:
	
		hc.trigTime= TIMER_getCount(timer1);

		/* Loop over SDSPs, sending a trigger to any which contained an error, and
		   checking to see that a good event arrives. */
		for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
			err[sdsp]= hc.dspInfo[sdsp].delErr;

			if (!(hc.pulsing & (1<<sdsp))) continue;  //SDSP is finished now.
			if (!err[sdsp]) continue; //Events are ok on this SDSP; don't bother.
			//L1-Ids were updated in resetModule, now just load them:
			writeRegisterDirect(CAL_L1_ID_0, 32, 0, hc.dspInfo[sdsp].L1id[0]);
			writeRegisterDirect(CAL_L1_ID_1, 32, 0, hc.dspInfo[sdsp].L1id[1]);
	
			if (hc.dspInfo[sdsp].paired) twin= hc.dspInfo[sdsp].pair;
			else                         twin= sdsp;

			writeRegisterDirect(DFLT_ROD_EVT_TYPE, 32, 0, sdsp);
			writeRegisterDirect(CRTV_ROD_EVT_TYPE, 32, 0, twin);
	
			if (!GET_RBIT(DIAGNOSTIC_REG, DR_USE_MDSP_MASK_LUT))
				writeRegisterDirect(FE_MASK_LUT_SELECT, 3, 0, sdsp);
			else
				switchLinkMasks(((DATA_LINK)+(COMMAND_LINK)), sdsp);
	
			//Send a new event:
			simTriggerArmM(sdsp);
			serialStreamOut(0, (TransData *) hc.ptr, hc.nSets);
		}

		//Wait until they've transmitted fully:
		t0= TIMER_getCount(timer1);
		status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt1= dt;
		do {
			delay(20); //2 us delay.
			status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt2= dt;
			ready= (icnt2 == icnt1); 
			icnt1= icnt2;
		} while ((!ready) && ((dt= delta_t(t0)) < HCTRL_TIMEOUT));
		cnt= delta_t(hc.trigTime);  icnt2= cnt>>4;
		delay(icnt2);

		nErrs= scanEventCheck(TRUE);
		updateScanStatistics();   //Report the new events/errors.

		++attempts;
		if (attempts >= MAX_RECOVER_ATTEMPTS) failed= TRUE;
	} while ((nErrs > 0) && (!failed));

	if (failed) {
		sprintf(genStr, "%s%d%s", "Problems with electronics- could not correct "
		                          "errors after ", MAX_RECOVER_ATTEMPTS,
		                          " attempts. Exiting.\n");
		newError(&returnCode, ELECTRONICS_ERROR, FATAL_ERR, "scanErrHandle",
		         genStr, __FILE__, __LINE__);

		for (sdsp=0; sdsp<N_SDSP; ++sdsp) addDspInfo(ERR_BUFF, sdsp);
	}

	return returnCode;
}

/************************************************************************************
 * updateScanStatistics
 *   
 *   synopsis: Updates the scan's statistics for the events which have arrived so
 *             far on the SDSPs. Events may still be en-route through the ROD
 *             while the routine executes-- it only cares about the events collected
 *             so far.
 ************************************************************************************/
void  updateScanStatistics(void) {
	register UINT32 a, b, c, cnt, sdsp, evt, err;

	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (!(hc.slvBits & (1<<sdsp))) continue;

		evt= hc.dspInfo[sdsp].seriesEvents[1] -hc.dspInfo[sdsp].seriesEvents[0];
		err= hc.dspInfo[sdsp].seriesErrors[1] -hc.dspInfo[sdsp].seriesErrors[0];

		//Since they're being added to the totals, update the old series counts:
		hc.dspInfo[sdsp].seriesEvents[0]+= (evt -err);
		hc.dspInfo[sdsp].seriesErrors[0]+= err;
		
		if      (sdsp == 0) a= HSSTAT_REG_0;
		else if (sdsp == 1) a= HSSTAT_REG_1;
		else if (sdsp == 2) a= HSSTAT_REG_2;
		else if (sdsp == 3) a= HSSTAT_REG_3;
	
		hc.dspInfo[sdsp].nErrors+= err;
		b= GET_RFLD(HSTATUS_REG_1, HSTAT1_NERR, HSTAT1_NERR_W);
		b+= err;
		ASGN_RFLD(HSTATUS_REG_1, HSTAT1_NERR, HSTAT1_NERR_W, b);

		cnt= hc.dspInfo[sdsp].updateCnt;

		b= hc.dspInfo[sdsp].nEvents;   b+= (evt -err);
		WRITE_REG(a, b);   hc.dspInfo[sdsp].nEvents= b;

		a= getSlvReg(sdsp, SDSP_HSTATUS_REG_3);
		c= hc.dspInfo[sdsp].avgProcTime; 

		hc.dspInfo[sdsp].avgProcTime= (cnt*c +a)/(cnt +1);

		++hc.dspInfo[sdsp].updateCnt;
	}

#if 0
		w= 	hc.dspInfo[sdsp].avgEvtLen;
		x= getSlvReg(sdsp, TRAPSTAT_REG_0);
		y= ++hc.dspInfo[sdsp].nGoodEvt;
		z= GET_VFLD(x, TSR0_EVT_WORD_CNT, TSR0_EVT_WORD_CNT_W);
	
		x= ((y-1)*w +z)/y;
	
		hc.evtLen= (UINT16) z;
		hc.dspInfo[sdsp].avgEvtLen= x;
#endif

}

/************************************************************************************
 * scanExtrapolate
 *   
 *   synopsis: Extrapolates using the previous event statistics to extend the duration
 *             of the event loop. This routine must be called after all the current
 *             statistics have been updated, as it may initiate a new trap series.
 ************************************************************************************/
typedef struct {
	UINT32  nTrig, seriesEvts, nEvts, nErr,
	        evtChk, del, newLoopMax, waitFlag;
} SdspLoopInfo;

typedef struct {
	UINT32 timeStamp, rep, origLoopMax, origWait,
	       loopMax, waitFlag, trigField, timeStamp2;

	SdspLoopInfo sdspLoop[N_SDSP];
} GenLoopInfo;

GenLoopInfo *genLoopInfo= (GenLoopInfo *) 0x02b00000;

void  scanExtrapolate(void) {
	register int sdsp, evt, del, rep= hc.repetitions;
	
	/* Max. possible # of triggers in each iteration of the trigger loop is
	   255 (0xff), since a bitfield (trigSent) which has 8 bits/sdsp is used
	   to count them. The scan task must give time for the main loop and any
	   other potential tasks/ primitive lists to run, and so it exits after
	   128 events max./iteration. (Note: the inefficiency introduced by
	   allowing the task to exit like this is negligible).
	   
	   The trigger loop maximum begins with 1 event, and then accelerates
	   quickly to the iteration maximum. Once the iteration maximum would
	   exceed the #repetitions/bin, the loop is scaled down to match. */
	if      (hc.trigLoopMax < 0x40)   hc.trigLoopMax<<= 3;
	else if (hc.trigLoopMax < 0x80)   hc.trigLoopMax<<= 1;
	if (hc.trigLoopMax > 0x80) hc.trigLoopMax= 0x80;
	hc.evtWait= FALSE;
	
	genLoopInfo->timeStamp=   TIMER_getCount(timer1);
	genLoopInfo->rep=         hc.repetitions;
	genLoopInfo->origLoopMax= hc.trigLoopMax;
	genLoopInfo->origWait=    hc.evtWait;

	for (sdsp=0; sdsp<N_SDSP; ++sdsp) {
		if (hc.dspInfo[sdsp].nTrig >= hc.repetitions) hc.pulsing &= ~(1<<sdsp);
		if (!(hc.pulsing & (1<<sdsp))) continue;  //SDSP is finished => don't care.

		if (hc.dspInfo[sdsp].seriesEvents[1]> 0x8000) newTrapSeries();

		evt= hc.dspInfo[sdsp].nTrig;
		del= rep -evt;
		if (del <= hc.trigLoopMax) {
			hc.trigLoopMax= del;
			hc.evtWait= TRUE;
		}

		genLoopInfo->sdspLoop[sdsp].nTrig=      hc.dspInfo[sdsp].nTrig;
		genLoopInfo->sdspLoop[sdsp].seriesEvts= hc.dspInfo[sdsp].seriesEvents[1];
		genLoopInfo->sdspLoop[sdsp].nEvts=      hc.dspInfo[sdsp].nEvents;
		genLoopInfo->sdspLoop[sdsp].nErr=       hc.dspInfo[sdsp].nErrors;
		genLoopInfo->sdspLoop[sdsp].evtChk=     evt;
		genLoopInfo->sdspLoop[sdsp].del=        del;
		genLoopInfo->sdspLoop[sdsp].newLoopMax= hc.trigLoopMax;
		genLoopInfo->sdspLoop[sdsp].waitFlag=   hc.evtWait;
	}

	genLoopInfo->loopMax=     hc.trigLoopMax;
	genLoopInfo->waitFlag=    hc.evtWait;
	genLoopInfo->trigField=   hc.pulsing;
	genLoopInfo->timeStamp2=  TIMER_getCount(timer1);
	if ((++genLoopInfo) >= ((GenLoopInfo *) 0x02c00000))  //wrap.
		genLoopInfo= (GenLoopInfo *) 0x02b00000;
}


/************************************************************************************
 * histoCtrlInit
 *   
 *   synopsis: Performs various initializations and checks needed before the histogram
 *             scan can begin. The routine reads the task parameters into a dedicated
 *     global array, performs several checks to ensure the integrity of the input,
 *     sets the ROD's running mode and sets a variety of variables which coordinate
 *     the data-taking among different groups of modules. A separate initialization
 *     routine (histoCtrlReady) is used to set up the SDSPs.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlInit(TaskInput *tsi) {
	INT32  returnCode= SUCCESS, errorCode;
	UINT32 i, j, type, t0;
	UINT8  sdsp, twin, set, k, nPoints;
	MDAT32 delta;

	t0= TIMER_getCount(timer1);
	setTaskState(HISTOGRAM_CTRL_TASK, TASK_INIT);

	/* Zero out the control structures Note that these structures contain several
	   variables which would naively be best left as local variables. However the DSP
	   has some limitations on the number of local variables which can be set up
	   before mysterious errors crop up (& before stack overflow too...) histoCtrl 
	   and scanTiming need a large number of variables; therefore it's safer to simply
	   place most of them inside a global structure. */
	setMem((UINT32 *) &hc, SIZEOF(MasterHistoCtrl), 0);
	setMem((UINT32 *) &scanStatus, SIZEOF(ScanStatus), 0);
	hc.scan= (ScanControl *) tsi;
	hc.status= (ScanStatus *) &scanStatus;
	hc.timing= (ScanTimingInfo *) &scanStatus.timing;
	hc.initTime = hc.time= t0;

	/* The trigger loop's max. time is set here; could imagine making this settable
	   via the task's priority is some way. */
	hc.taskTimeMax= 5000000; //.5 x 10^6 us 		

	//Get the task's control bits in the diagnostic register:
	getTaskCtrlBits();

	hc.extSetup= hc.scan->rodSetup.extSetup; 

	if (hc.extSetup & EXT_SETUP_RCF) {
		if (rodModeCfg[0].rodMode == CALIBRATION_MODE) hc.dataPath= rodModeCfg[0].sim;
		else                                           hc.dataPath= DATA_PATH_NORMAL;
	}
	else { //Set up calibration mode:
		/* Set the ROD's running mode to calibration, preserving the simulation mode
		   modifier if it is present; for other modifiers, the RCF must be set up
		   externally. The old mode is stored. */
		hc.dataPath= hc.scan->rodSetup.dataPath; 
		if (hc.dataPath == DATA_PATH_NORMAL)  i= CALIBRATION_MODE +INMEM_EVT_CAPTURE_MODE;
		else                                  i= CALIBRATION_MODE +SIMULATION_MODE;

		i+= CALIBRATION_SLINK_OVERRIDE_MODE;
		rodMode(i, STORE_MODE, TRUE, 0x1000, 0, 1, TRUE);
	}

	writeRegister(EFB_CMND_0, 1, EFB_GROUP_COUNTER_EN_O, 1);	 //Enable EFB counters.


//dpsf: place all in histoCtrlSlvSetup, load variables appropriately
	//SDSPs which will be  histogramming (bitfield):
	hc.slvBits= hc.scan->dspDist.slvBits;

	//hc.timing->init.dt[0]= delta_t(hc.initTime);
	hc.time= TIMER_getCount(timer1);

	hc.cfgSet=   hc.scan->general.cfgSet;
	if (hc.cfgSet >= N_MODULE_CONFIG_SETS) {
		sprintf(genStr, "%s%d%s","Invalid configuration set (max= ",
			                     (N_MODULE_CONFIG_SETS-1), ").\n");
		newError(&returnCode, PRIM_PARAMETER_ERROR, FATAL_ERR, "histoCtrlInit",
		         genStr, __FILE__, __LINE__);
		return returnCode;
	}
	hc.moduleCfg=  moduleConfigSet[hc.cfgSet];

	hc.port= hc.scan->serial.port;
	//If requested, reset modules using the PHYSICS config set:
	if (hc.scan->reset.moduleInit) {
		//type= (1<<CONFIG_MODULE_ALL);
		errorCode= sendConfigSet(hc.port,  hc.mod8, ALL_CHIPS,  NORMAL_CONFIG_LOOP,
		                         TRUE, PHYSICS_CONFIG_SET, MODULE_GROUP_ALL,
		                         CONFIG_MODULE_ALL, FALSE, TRUE);
	}

	/* Update the Module configuration structures to use the scan-specific information.
	   This sets FE registers on the modules into send-mask mode, for example. The
	   PHYSICS configuration can be just copied into the SCAN slot when preparing
	   the scan.  */
//	for (i=0; i<N_MODULES; ++i) {
//	if ((hc.moduleCfg[i].present)&&(hc.moduleCfg[i].active)) {
//		errorCode= updateModuleStructure(&hc.moduleCfg[i], hc.scan);
//	}}

	if (hc.manual) scanWait(SCAN_INIT_CFG, 0);

	//hc.timing->init.dt[1]= delta_t(hc.time);
	hc.timing->init.dt_cfg= cfgTiming.total;
	//hc.timing->init.dt_cfg_setup= cfgTiming.setup;

	hc.time= TIMER_getCount(timer1);

	//hc.timing->init.dt[2]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	//Stage configuration mask is in basic register (for sendCfg):
	hc.stageCfg= CONFIG_MODULE_BASIC;

	//Determine which register contains the scan parameter:
	for (i=0; i<2; ++i) {
		hc.cfgReg[i]= hc.scan->general.scanParameter[i];
		hc.sendCfg[i]= TRUE;   hc.singleCfgVar[i]= FALSE;

		switch(hc.cfgReg[i]) {
			case SCAN_NONE:
			case ST_ELAPSED:
				hc.dataType[i]= CONFIG_NONE;    hc.sendCfg[i]= FALSE;
				break;

			case ST_VTHR:    case ST_VCAL:    case ST_STROBE_DELAY:    case ST_PREAMP:
			case ST_SHAPER:

			case ST_QTHR:    case ST_QCAL:    case ST_TTHR:

			case ST_TARGET:

			case ST_ROLE:
				hc.dataType[i]= CONFIG_MODULE_BASIC;
				break;

			case ST_TRIM:
				hc.dataType[i]= CONFIG_MODULE_TRIM; //dpsf: CONFIG_MODULE_SCAN_TRIM;
				break;
			
			case ST_MASK:          case ST_NMASK:
				hc.dataType[i]= CONFIG_MODULE_BASIC; //dpsf: CONFIG_MODULE_MASK;
				break;

			case ST_BYPASS:  case ST_TOKEN:
				hc.dataType[i]= CONFIG_MODULE_BASIC;
				hc.singleCfgVar[i]= TRUE;
				break;

			/* ROD trigger delay(s): */
			case ST_TRIG_DELAY1:   case ST_TRIG_DELAY2:
				hc.dataType[i]= CONFIG_ROD_TRIGGER;       hc.sendCfg[i]= FALSE;
				break;

			/* BOC parameters: */
			case ST_RX_DELAY:        case ST_RX_DELAY0:       case ST_RX_DELAY1:
			case ST_RX_THRESHOLD:    case ST_RX_THRESHOLD0:   case ST_RX_THRESHOLD1:
			case ST_TX_CURRENT:      case ST_TX_MARKSPACE:    case ST_TX_DELAY:
			case ST_TX_COARSE:       case ST_TX_FINE:
				hc.dataType[i]= CONFIG_BOC;               hc.sendCfg[i]= FALSE;
				hc.singleCfgVar[i]= TRUE;
				break;

			default:
				hc.dataType[i]= CONFIG_MODULE_ALL;
				break;
		}
	}

	for (i=0; i<2; ++i) {
		hc.dataPtr[i]=   hc.scan->general.dataPtr[i];
		hc.nBins[i]=     hc.scan->general.nBins[i];		

		if (hc.nBins[i] > HISTOGRAM_BIN_MAX) {
			sprintf(genStr, "%s%d%s","Maximum number of bins is ",
				                     HISTOGRAM_BIN_MAX, ".\n");
			newError(&returnCode, PRIM_PARAMETER_ERROR, FATAL_ERR, "histoCtrlInit",
			         genStr, __FILE__, __LINE__);
			return returnCode;
		}

		for (j= 0; j< HISTOGRAM_BIN_MAX; ++j) {varRange[i][j]= 0; }
	}
	hc.repetitions=    hc.scan->general.repetitions;

	for (i=0; i<2; ++i) {
//dpsf: temp	if (hc.cfgReg[i] != SCAN_NONE) {
		if (hc.scan->general.uniformPoints[i]) {
			nPoints= hc.scan->general.nBins[i] -1;
			if (nPoints <= 0) nPoints= 1;
			delta= (hc.scan->general.scanEnd[i] -hc.scan->general.scanStart[i])/nPoints;

			for (j= 0; j< hc.nBins[i]; ++j) {
				varRange[i][j]= hc.scan->general.scanStart[i] +j*delta;
			}
		}

		else {  /* read in from pre-stored locations in memory */
//dpsf: need new var. distinguishing between dual ranges & 2D scan.
//dpsf: scanDim= 1 or 2.
#if 1
			for (j= 0; j< hc.nBins[0]; ++j) {
#else
			for (j= 0; j< hc.nBins[i]; ++j) {
#endif
				varRange[i][j]= *(hc.dataPtr[i] +j);
			}
		}
	} //}

	//hc.timing->init.dt[3]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	/* Set initial variable-loading settings for the initial pass through NEWBIN: */
	hc.twoDScan= ((hc.cfgReg[1] != SCAN_NONE) && (hc.scan->general.nBins[1] > 0));

	hc.stageAdvance= hc.binAdvance[1]= FALSE;
	if (hc.scan->general.maskMode == SCAN_STAGED) {
		hc.stageAdvance= TRUE;
		hc.status->currentMaskStage= 0;
	}

	hc.binAdvance[0]= TRUE;
	hc.status->currentBin[0]= hc.status->currentBin[1]= 0;
	if (hc.twoDScan) hc.binAdvance[1]= TRUE;

	//hc.timing->init.dt[4]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);


	/* Modules are arranged into groups; each group has a (possibly) different
	   SDSP destination and function (e.g. Some groups can generate interference).
	   Group mappings to the serial ports and SDSPs must be unique (a group's
	   events arrive on a single DSP, and command data is sent to it from a
	   single serial port). */
	hc.defGroups= hc.scan->dspDist.definedGroups;
	for (i=0; i<2; ++i) {
		hc.groupSPMap[i]=     hc.scan->dspDist.groupSPMap[i]    & hc.defGroups;
		hc.groupRangeMap[i]=  hc.scan->dspDist.groupRangeMap[i] & hc.defGroups;
	}

	if (  ((hc.groupRangeMap[0] ^ hc.groupRangeMap[1]) != hc.defGroups)
	    ||((hc.groupSPMap[0]    ^ hc.groupSPMap[1])    != hc.defGroups) ) {

		sprintf(genStr, "%s", "Each group must map onto a unique "
			                  "range and serial port.\n");
		newError(&returnCode, PARAM_ERROR, FATAL_ERR, "histoCtrlTask",
		         genStr,  __FILE__, __LINE__);
		return returnCode;
	}

	if (  ((hc.port == SP0) && (hc.groupSPMap[1]))
	    ||((hc.port == SP1) && (hc.groupSPMap[0])) ) {
		newError(&returnCode, PARAM_ERROR, FATAL_ERR, "histoCtrlTask",
		         "Illegal group serial port mapping.\n",  __FILE__, __LINE__);
		return returnCode;
	}

	for (i=0; i<2; ++i) for (j=0; j<2; ++j) hc.dspPair[i][j]= 0xff;
	hc.nDspPairs= hc.scan->dspDist.nDspPairs;
	for (i=0; i<hc.scan->dspDist.nDspPairs; ++i) {
		for (j=0; j<2; ++j) hc.dspPair[i][j]= hc.scan->dspDist.dspPair[i][j] & 0x3;
	}

	for (j=sdsp=0; sdsp<N_SDSP; ++sdsp) {
		hc.groupDSPMap[sdsp]= hc.scan->dspDist.groupDSPMap[sdsp] & hc.defGroups;		
		j+= (hc.groupDSPMap[sdsp] & hc.defGroups);
	}
	if (j != hc.defGroups) {
		sprintf(genStr, "%s","Each group must map onto a unique DSP.\n");
		newError(&returnCode, PARAM_ERROR, FATAL_ERR, "histoCtrlTask",
		         genStr,  __FILE__, __LINE__);
		return returnCode;
	}

	/* Initialize the internal trigger sequences, and count the number of commands
	   and triggers in them. Also initialize the internal ECR/BCR lists and null
	   lists. */
	for (i=j=k=0; k<N_CMD_LIST_CMDS; ++k) {
		hc.triggerSequence[0].cmd[k]=  NO_CMD;
		hc.triggerSequence[1].cmd[k]=  NO_CMD;

		hc.triggerSequence[0].data[k]=  0;
		hc.triggerSequence[1].data[k]=  0;

		nullCmdList.cmd[k]=  NO_CMD;  //dpsf patch.
		nullCmdList.data[k]= 0;
	}

	/* The trigger can be either built using the input delay between calibration
	   pulse and L1A or, for more specialized applications, read in from two input
	   command lists. If either of the two scan dimensions is on the calibration-
	   pulse -> L1A delay, the trigger streams are implicitly assumed to be
	   calculated from the structure. */
	if (  (hc.scan->general.scanParameter[0] == ST_TRIG_DELAY1)
	    ||(hc.scan->general.scanParameter[1] == ST_TRIG_DELAY1)
	    ||(hc.scan->general.scanParameter[0] == ST_TRIG_DELAY2)
	    ||(hc.scan->general.scanParameter[1] == ST_TRIG_DELAY2) )
		hc.scan->serial.calcFromStruct= TRUE;    

	if (hc.scan->serial.calcFromStruct) {
		hc.status->currentDelay= hc.scan->serial.calL1ADelay;
		buildScanTrigger(TRUE);  /* Initialize the command lists */
	}
	else {
		if ((hc.port == SP0) || (hc.port == SP_BOTH))
			hc.triggerSequence[0]= hc.scan->serial.triggerSequence[0];		
		if ((hc.port == SP1) || (hc.port == SP_BOTH))
			hc.triggerSequence[1]= hc.scan->serial.triggerSequence[1];		
	}

	for (i=j=k=0; k<N_CMD_LIST_CMDS; ++k) {
		if (hc.triggerSequence[0].cmd[k] !=  NO_CMD) ++i;
		if (hc.triggerSequence[1].cmd[k] !=  NO_CMD) ++j;
	}
	//if (i == 0) hc.triggerSequence[0]= (CmdList *) NULL;
	//if (j == 0) hc.triggerSequence[1]= (CmdList *) NULL;

	/* Count the # of trigger sequences with an L1A in them. This will be used
	   later to determine the modules' trigger reception counts. */
	for (i=0; i<2; ++i) {
		for (k=0; k<N_CMD_LIST_CMDS; ++k) {
			if (hc.triggerSequence[i].cmd[k] ==  L1_TRIGGER) ++hc.SPTriggers[i];
			if (hc.triggerSequence[i].cmd[k] ==  SOFT_RESET) hc.SPEcr[i]= TRUE;
		}
	}
	hc.triggerPairs= FALSE;
	if ((hc.SPTriggers[0] > 0) && (hc.SPTriggers[1] > 0))  hc.triggerPairs= TRUE;

	//Build the L1-ID reset flag:
	hc.L1idReset= (  (hc.dataPath == DATA_PATH_INMEM)
	               ||((hc.SPEcr[0]) || (hc.SPEcr[1]))  );

	/* Create maps from the bitfields: first fill in the groups' structures;
	   then loop over the slave DSPs, using the contributing groups'
	   information to calculate the ncessary masks & other DSP information. */
	for (i=0; i<4; ++i)
		setLinkMasks(0, SP_BOTH, ((DATA_LINK)+(COMMAND_LINK)), INIT_MASK, i);


	for (i= 0; i < N_MODULE_GROUPS; ++i) {
		for (j= 0; j < 2; ++j) {
			if (hc.groupSPMap[j] & (1<<(i)) )  hc.groupInfo[i].serialPort= j;
			if (hc.groupRangeMap[j] & (1<<(i)) )  hc.groupInfo[i].rangeList= j;
		}

		/* Loop over modules in the configSet. If a module is present and in
		   that module group, add it to the group's module map. Note that the
		   extra "off-ROD" modules must have their data links turned off. */
		for (k=0; k<N_TOTMODULES; ++k) {
		if ((hc.moduleCfg[k].present) && (hc.moduleCfg[k].groupId == i)) {
			if      (k < 32) hc.groupInfo[i].validModules[0]|= (1<<(k));
			else             hc.groupInfo[i].validModules[1]|= (1<<(k-32));
		}} 
	}    /* loop over module groups */


	//hc.timing->init.dt[5]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);



	for (sdsp= 0; sdsp < N_SDSP; ++sdsp) {
		hc.dspInfo[sdsp].dspNum= sdsp;

		/* The mask allows the trigger loop to make quick bitfield-based decisions
		   about issuing triggers to SDSPs: */
		hc.dspInfo[sdsp].mask= 0xff;  //Default "inactive" mask.

		if (!(hc.slvBits & (1<<sdsp))) continue;
		for (j= 0; j < N_MODULE_GROUPS; ++j) hc.dspInfo[sdsp].groupId[j]= 0xff;

		hc.dspInfo[sdsp].active= TRUE;
		hc.dspInfo[sdsp].paired= FALSE;
		hc.dspInfo[sdsp].mask= (1<<sdsp);  //Default active mask

		if (sdsp == hc.dspPair[0][0]) {
			twin= hc.dspInfo[sdsp].pair= hc.dspPair[0][1];
			hc.dspInfo[sdsp].paired= TRUE;
			hc.dspInfo[sdsp].mask= (1<<sdsp) | (1<<twin); //Active/paired mask
		}
		else if (sdsp == hc.dspPair[1][0]) {
			twin= hc.dspInfo[sdsp].pair= hc.dspPair[1][1];
			hc.dspInfo[sdsp].paired= TRUE;
			hc.dspInfo[sdsp].mask= (1<<sdsp) | (1<<twin); //Active/paired mask
		}
		else if (sdsp == hc.dspPair[0][1]) {
			hc.dspInfo[sdsp].pair= hc.dspPair[0][0];
			hc.dspInfo[sdsp].active= FALSE;
			hc.dspInfo[sdsp].paired= TRUE;
			hc.dspInfo[sdsp].mask= 0x0f;  //Default passive mask (but in scan)
		}
		else if (sdsp == hc.dspPair[1][1]) {
			hc.dspInfo[sdsp].pair= hc.dspPair[1][0];
			hc.dspInfo[sdsp].active= FALSE;
			hc.dspInfo[sdsp].paired= TRUE;
			hc.dspInfo[sdsp].mask= 0x0f;  //Default passive mask (but in scan)
		}

		if   (hc.dspInfo[sdsp].active == TRUE) set= sdsp;
		else                                   set= hc.dspInfo[sdsp].pair;


		for (j= 0; j < N_MODULE_GROUPS; ++j) {
		if (!(hc.groupDSPMap[sdsp] & (1<<j))) continue; //Group does not contribute.

			for (k=0; k<N_TOTMODULES; ++k) {
			if ((hc.moduleCfg[k].present) && (hc.moduleCfg[k].groupId == j)) {
				setLinkMasks(k, hc.groupInfo[j].serialPort,
				             ((DATA_LINK_PLAY) +(COMMAND_LINK_ON)), UPDATE_MASK, set);
			}}

			hc.dspInfo[sdsp].validModules[0]|= hc.groupInfo[j].validModules[0];
			hc.dspInfo[sdsp].validModules[1]|= hc.groupInfo[j].validModules[1];
			for (i=0; i<32; ++i) {
				if (hc.dspInfo[sdsp].validModules[0] & (1<<i)) ++hc.dspInfo[sdsp].nModules;
				if (hc.dspInfo[sdsp].validModules[1] & (1<<i)) ++hc.dspInfo[sdsp].nModules;
			}

			k= hc.groupInfo[j].serialPort;
			hc.dspInfo[sdsp].ports|= (k+1);


			/*    Group event & L1id counters are pointers which point to the
			   corresponding SDSP's counter. If the MDSP encounters problems
			   while sending an event to that slave, all the modules on both
			   serial ports receive a reset. Different counters are used for the
			   number of triggers successfully sent (and events received) for
			   the modules on each serial port.
			   
			      If events are distributed according to module groups, each module
			   contributes events to one DSP. If router distribution is being used,
			   every module will send events to all SDSPs; in this case, the
			   counters for SDSP 0 serve as global counters. */
			hc.groupInfo[j].L1id= &hc.dspInfo[sdsp].L1id[k];
			//hc.groupInfo[j].nEvents= &hc.dspInfo[sdsp].nEvents[k];
			//hc.groupInfo[j].nErrors= &hc.dspInfo[sdsp].nErrors[k];

			++hc.dspInfo[sdsp].nContribGroups;
			hc.dspInfo[sdsp].groupId[hc.dspInfo[sdsp].nContribGroups-1]= j;
		}  //Groups contibuting to this DSP


		if (hc.dspInfo[sdsp].ports & 1) ++hc.dspInfo[sdsp].nPorts;
		if (hc.dspInfo[sdsp].ports & 2) ++hc.dspInfo[sdsp].nPorts;

		//one event only for now
		if (hc.dspInfo[sdsp].nPorts == 2) {
			sprintf(genStr, "%s","Each DSP must receive a single event per trigger.\n");
			newError(&returnCode, PARAM_ERROR, FATAL_ERR, "histoCtrlTask",
			         genStr,  __FILE__, __LINE__);
			return returnCode;
		}

	}  //Gather DSP information

	/* If DSP pairing is being used, check that the two DSPs each use a different
	   serial port. The 1st member of the pair must use SP 0 and the 2nd member
	   must use SP 1. */
	for (i=0; i<hc.nDspPairs; ++i) {
		if (  (hc.dspInfo[hc.dspPair[i][0]].nPorts > 1)
		    ||(hc.dspInfo[hc.dspPair[i][1]].nPorts > 1)
		    ||(hc.dspInfo[hc.dspPair[i][0]].ports
		       == hc.dspInfo[hc.dspPair[i][1]].ports)
		    ||(hc.dspInfo[hc.dspPair[i][0]].ports != 1)  ) {
			sprintf(genStr, "DSP pairing error.\n");
			newError(&returnCode, PARAM_ERROR, FATAL_ERR, "histoCtrlTask",
			         genStr,  __FILE__, __LINE__);
			return returnCode;
		}
	}


	//hc.timing->init.dt[6]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	//hc.timing->init.dt[7]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	/* Pre-compute the mask-set deltas for each SDSP if the masks will be changing;
	   note that we only care about the serial port command link masks and the dynamic
	   mode bits (DATA_LINK). The link configuration (on or off) is left unchanged by
	   these mask sets to increase switching speed. This is done in a separate SDSP
	   loop to allow paired SDSPs to be linked together. */
	if (GET_RBIT(DIAGNOSTIC_REG, DR_USE_MDSP_MASK_LUT)) {
		for (sdsp= 0; sdsp < N_SDSP; ++sdsp) {
			setLinkMasks(0, SP_BOTH,
			             ((DATA_LINK) +(COMMAND_LINK)), COMPUTE_MASK_DELTAS, sdsp);
		}
	
		setLinkMasks(0, SP_BOTH, ((DATA_LINK) +(COMMAND_LINK)), COMPUTE_MASK_DELTAS,
		             MASK_SET_INIT);
	}

	setLinkMasks(0, SP_BOTH, ((DATA_LINK) +(COMMAND_LINK)), SET_MASK, MASK_SET_INIT);

	hc.slvRep= hc.repetitions;

	hc.reTransmitCnt= 0;

	//hc.timing->init.dt[8]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);


	/* Get pointers to & initialize the sp buffers */
	hc.cmdBuffer[0]= &cmdBuffer[0];
	hc.cmdBuffer[1]= &cmdBuffer[1];
	initCmdBuffer(hc.port);


	/* Initialize the post-configuration command list structure. This is used
	   later to re-enable data taking (the configuration commands will throw the
	   ABCD chips into send-config mode), and issue a BC reset (to ensure that
	   the modules & ROD are in synch). */
	for (i=0; i<N_CMD_LIST_CMDS; ++i) {
		hc.postConfigCmdList.cmd[i]=  NO_CMD;
		hc.postConfigCmdList.data[i]= 0;
	}
	
	hc.cmdListPtr[0]= hc.cmdListPtr[1]= &hc.postConfigCmdList;
	if      (hc.port == SP0) hc.cmdListPtr[1]= NULL;
	else if (hc.port == SP1) hc.cmdListPtr[0]= NULL;

#if 1
	errorCode= initSerialStreams();
	if (errorCode != SUCCESS) {
		addError(&returnCode, errorCode, "histoCtrlInit", "initSerialStreams",
		         __FILE__, __LINE__);
		return returnCode;
	}
#endif

	//hc.timing->init.dt[9]= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	hc.timing->init.total = delta_t(hc.initTime);
	hc.time= TIMER_getCount(timer1);

	if (hc.manual) scanWait(SCAN_END_INIT, 0);
	//return returnCode;

#if 1
//dpsf: temp. belongs in sdsp setup really.
	/* override the task manager's event filter on the slaves temporarily, and set
	   the slaves' trap cmd-stat register to indicate that triggers will be pending */
	for (sdsp= 0; sdsp< N_SDSP; ++sdsp) {
		if (!(hc.slvBits & (1<<sdsp))) continue; 
		setSlvRegBit(sdsp, COMMAND_REG_0, CR_HISTO_OVERRIDE); 
		//setSlvReg(sdsp, TRAP_CMD_STAT, (0x0 | (1<<TCSR_ISR_PENDING)) );
	}

	do {
		errorCode= chkSlaves(hc.slvBits, &hc.ready, SDSP_HSTATUS_REG_0, HSR0_SLV_RDY);
	} while (((hc.interval= delta_t(hc.time)) < 5000) && (hc.slvBits != hc.ready));

	if (FATAL(errorCode)) {
		addError(&returnCode, errorCode, "histoCtrlInit", "chkSlaves",
		         __FILE__, __LINE__);
		return returnCode;
	}
	else if (hc.ready != hc.slvBits) {
		sprintf(genStr, "%s%s%d%s","The histogramming task is not running on ",
			        " one of the slaves, or it is not under master control.\n");
		newError(&returnCode, SLAVE_NOT_READY, FATAL_ERR, "histoCtrlInit",
		 genStr,  __FILE__, __LINE__);
		return returnCode;
	}

#endif

	return returnCode;
}


/************************************************************************************
 * initSerialStreams
 *   
 *   synopsis: Initializes & builds all the various (static or nearly so) serial
 *             streams which are needed during a scan. These are of various types
 *      which: control resetting the modules, configuring the module's mask stages,
 *      configuring the module's scan variables, and send the trigger sequence.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 initSerialStreams() {
	INT32 returnCode= SUCCESS, status;
	return returnCode;
}

#if 0
	UINT32 *streamData[2], i, j, k, mask[4];

	/* Stage masks:  Normally, staging is treated by broadcasting the current mask
	   stage (pre-computed, for all chips) to all modules; this is quite fast. If
	   any dead or bad channels are required to be masked off, the procedure is
	   modified by sending the above serial stream only to those modules with
	   normal (i.e. unblemished) masks, and then individually treating any modules
	   with chips having dead channels using pre-computed serial streams tailored
	   for that module. Finally, any off-ROD modules (using redundant command lines 
	   and the data arrives on another ROD) are configured. */

	//Initialize the command buffers:
	for (i=0; i<4; ++i) {
		if ((hc.port == SP_BOTH) || (hc.port == SP0)) k= 0;
		else                                          k= 1;

		streamData[k]=   &stageStream[i][0];
		streamData[k^1]= NULL;
		hc.mod8[0]= hc.mod8[1]= NO_MODULE;

		initCmdBuff(&stageCmdBuff[i], streamData, &stageIlBuff[i][0], 8,
		            hc.port, hc.mod8);

		for (j=0; j<4; ++j) mask[j]= 0x11111111<<i;  //Form the channel mask.
		status= addMaskCmd(&stageCmdBuff[i], ALL_CHIPS, mask);
	}


	return returnCode;
}


	if (hc.stageMode == 0) {
		stageCmdBuff[i].cmdMask
	}



	setLinkMasks(0, SP_BOTH, COMMAND_LINK, SET_MASK, MASK_SET_INIT,
		        (UINT8) GET_RBIT(DIAGNOSTIC_REG, DR_USE_ROD_MASK_LUT), 0);

	setLinkMasks(0, SP_BOTH, COMMAND_LINK, INIT_MASK, MASK_SET_WORK0,
		        (UINT8) GET_RBIT(DIAGNOSTIC_REG, DR_USE_ROD_MASK_LUT), 0);

	setLinkMasks(0, SP_BOTH, COMMAND_LINK, STORE_MASK, MASK_SET_WORK0,
		        (UINT8) GET_RBIT(DIAGNOSTIC_REG, DR_USE_ROD_MASK_LUT), 0);
		        




		for (i=0; i<N_MODULES; ++i) {
			if (hc.configSetPtr[i].normalMasks != 0xf)
		}
	}

 

}
#endif


/************************************************************************************
 * histoCtrlSdspSetup
 *   
 *   synopsis: If requested, starts up and coordinates all the slave's data taking
 *             engines and histogramming tasks, then loops while waiting for the SDSPs
 *     to indicate that they are ready for histogramming to begin. The task must cede
 *     control to the main loop so that the inter-DSP list state machines can run.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlSdspSetup(TaskInput *tsi) {
	INT32  returnCode= SUCCESS, errorCode[5];
	static UINT8 slvBits, sdsp;
	static UINT32 initTime;
	UINT32  i, test, lerr, lfatal;

	return 0;
}
#ifdef COMMENTED
	if (hc.setupState == HCSETUP_INIT) {
		setTaskState(HISTOGRAM_CTRL_TASK, HCTRL_SDSP_SETUP);
		slvBits= hc.slvBits;   

		/* If any of the following 4 fields are not set, then we must create a
		   list and send it to the SDSPs via an inter-DSP list. Otherwise, we
		   simply check that the SDSPs have the histogramming task running on
		   them. */
		test= (EXT_SETUP_ETS +EXT_SETUP_SET +EXT_SETUP_HISTO  +EXT_SETUP_HTASK);
		if ((hc.extSetup & test) != test) hc.setupState= HCSETUP_LISTMAKE;
		else                              hc.setupState= HCSETUP_READY;

		/* Prepare the copies of the event trap structure for the router & SDSPs,
		   and set up the router: */
		if (!(hc.extSetup & EXT_SETUP_ETS)) {
			setMem((UINT32 *) &hc_ets_i, SIZEOF(hc_ets_i), 0);

			setMem((UINT32 *) &trapParams[0], SIZEOF(RouterTrapParams), 0);
			setMem((UINT32 *) &trapParams[1], SIZEOF(RouterTrapParams), 0);

			hc_ets_i.nEvents= COLLECT_FOREVER;
			hc_ets_i.releaseFrames= TRUE;
			hc_ets_i.permitBackPressure= TRUE;
			hc_ets_i.dataMode= FALSE;
			hc_ets_i.sLink= FALSE;
			hc_ets_i.format= TRAP_FMT_NORMAL;
			hc_ets_i.trapStray= TRUE;

//dpsf: ugly. fix.
			trapParams[0].config=        hc_ets_i.trapConfig[0]=    TRAP_CFG_ROD_EVT;
			trapParams[0].exclusionFlag= hc_ets_i.trapExclusionFlag[0]= FALSE;
			trapParams[0].function=      hc_ets_i.trapFunction[0]=  TRAP_FXN_HISTOGRAM;
			trapParams[0].modulus=       hc_ets_i.trapModulus[0]=   1;
			trapParams[0].remainder=     hc_ets_i.trapRemainder[0]= 0;

			//Set up the router for ROD events:
			for (sdsp= 0; sdsp<N_SDSP; ++sdsp) {
				if (!(slvBits & (1<<sdsp))) continue;
				trapParams[0].match= sdsp;
				setupRouter(sdsp, hc_ets_i.permitBackPressure, hc_ets_i.dataMode,
				            hc_ets_i.sLink, hc_ets_i.format, trapParams);
			}
		} //ETS setup.

		//SET has no input structure.

		if (!(hc.extSetup & EXT_SETUP_HISTO)) {
			setMem((UINT32 *) &hc_hsetup_i, SIZEOF(hc_hsetup_i), 0);
			
			hc_hsetup_i.base= (UINT32 *) DEFAULT;
			hc_hsetup_i.binSize= HISTOGRAM_16BIT;
			hc_hsetup_i.dataType[0]= hc.dataType[0];

			hc_hsetup_i.moduleRangeMap[0][0]= 0xffffffff;
			hc_hsetup_i.moduleRangeMap[0][1]= 0xffffffff;

			hc_hsetup_i.nBins= hc.nBins[0];

			hc_hsetup_i.opt[0]= TRUE;  //Occupancy
			hc_hsetup_i.opt[1]= TRUE;  //Time-slice
			hc_hsetup_i.opt[2]= 0;     //No TOT.
			hc_hsetup_i.opt[3]= hc.scan->trigger.accepts;

			//Later we will add the range lists:
			hc_hsetup_i.xPtr[0]= hc_hsetup_i.xPtr[1]= (MDAT32 *) DEFAULT;
		}

		if (!(hc.extSetup & EXT_SETUP_HTASK)) {
			setMem((UINT32 *) &hc_htask_i, SIZEOF(hc_htask_i), 0);

			hc_htask_i.completionFlag= TRUE;
			hc_htask_i.priority= 1;
			hc_htask_i.taskType= HISTOGRAM_TASK;
			hc_htask_i.taskRevision= R_HISTOGRAM_TASK;
			hc_htask_i.taskStruct.histogramTaskIn.nEvents= COLLECT_FOREVER;
			hc_htask_i.taskStruct.histogramTaskIn.controlFlag= MASTER_HREG;
		}

		hc.setupState= HCSETUP_LISTMAKE;
		return WAITING_FOR_SLAVES;
	}

	/* Build the primitive list to be sent to the SDSPs. The list is re-built for
	   each SDSP. The next SDSP will be handled after the list has been processed &
	   checked for the current SDSP. */
	else if (hc.setupState == HCSETUP_LISTMAKE) {
		/* get the first non-zero SDSP bit from the bitfield */
		for (i= 0; i< N_SDSP; ++i) {
			if (!(slvBits & 1<<i)) continue; 
			else {sdsp= i; break; }
		}

		checkSdspStatus(sdsp, 3, &returnCode, "histoCtrlSdspSetup");
		if (FATAL(returnCode)) return returnCode;

		if (!(hc.extSetup & EXT_SETUP_ETS)) {
			hc_ets_i.trapMatch[0]=         sdsp;
			errorCode[0]= addMessage(IDSP_SEND, PRM, EVENT_TRAP_SETUP,
			                      R_EVENT_TRAP_SETUP,SIZEOF(EventTrapSetupIn),
			                      (void *) &hc_ets_i);
		}

		if (!(hc.extSetup & EXT_SETUP_SET)) {
			errorCode[1]= addMessage(IDSP_SEND, PRM, START_EVENT_TRAPPING,
			                      R_START_EVENT_TRAPPING, 0, NULL);
		}

		if (!(hc.extSetup & EXT_SETUP_HISTO)) {
			hc_hsetup_i.validModules[0]= hc.dspInfo[sdsp].validModules[0];
			errorCode[2]= addMessage(IDSP_SEND, PRM, HISTOGRAM_SETUP,
			                      R_HISTOGRAM_SETUP, SIZEOF(HistogramSetupIn),
			                      (void *) &hc_hsetup_i);
		}

		if (!(hc.extSetup & EXT_SETUP_HTASK)) {
			errorCode[3]= addMessage(IDSP_SEND, PRM, START_TASK,
			                      R_START_TASK, SIZEOF(StartTaskIn),
			                      (void *) &hc_htask_i);
		}

		for (i=0; i<4; ++i) {
			if (errorCode[i] != SUCCESS) {
				addError(&returnCode, errorCode[i], "histoCtrlSdspSetup",
				         "SDSP list construction", __FILE__, __LINE__);
				if (FATAL(returnCode)) return returnCode;
			}
		}

		//addListWrapper(IDSP_SEND, PRM, 0, 0,0,0,0);
		initiateListSendSM(sdsp);

		initTime= TIMER_getCount(timer1);
		hc.setupState= HCSETUP_LISTWAIT;
		return WAITING_FOR_SLAVES;
	} 

	/* Wait while SDSP processes list */
	else if (hc.setupState == HCSETUP_LISTWAIT) {
		if (SLAVE_IDSP_LIST_BUSY) {
			if (delta_t(initTime) > HCTRL_TIMEOUT) {
				sprintf(genStr, "%s%d%s", "Timed out while waiting for SDSP #",
				                sdsp, " to finish processing IDSP primitive list.\n");
				newError(&returnCode, TIMEOUT_ERR, FATAL_ERR,
				         "histoCtrlSdspSetup", genStr,  __FILE__, __LINE__);

				hc.setupState= HCSETUP_INIT;
				return returnCode;
			}
		}
		else {hc.setupState= HCSETUP_LISTTEST; }
		
		return WAITING_FOR_SLAVES;
	}

	/* Check that the list completed without errors, and go on to the next SDSP,
	   if necessary. */
	else if (hc.setupState == HCSETUP_LISTTEST) {
		if ((lerr= getIntrDspErr()) || (lfatal= getIntrDspFatal())) { //SDSP problem?
			sprintf(genStr, "%s%d%s%d%s%d%s", "The inter-DSP list failed for SDSP #",
			                sdsp, " (err, fatal err ==  ",
			                lerr, " ", lfatal, ").\n");
			newError(&returnCode, PRIMLIST_ERR, FATAL_ERR,
			         "histoCtrlSdspSetup", genStr,  __FILE__, __LINE__);

			hc.setupState= HCSETUP_INIT;
			return returnCode;
		}

		slvBits^= 1<<sdsp;   /* process next slave */
		if (slvBits != 0) {hc.setupState= HCSETUP_LISTMAKE; }
		else              {hc.setupState= HCSETUP_READY; }
		return WAITING_FOR_SLAVES;
	}

	/* If all lists have been sent, go on to a final check of all enabled SDSPs and
	   then exit. */
	else if (hc.setupState == HCSETUP_READY) { 
		hc.setupState= HCSETUP_INIT;

		/* Set the slaves' trap cmd-stat register to indicate that triggers will be
		   pending */
		for (sdsp= 0; sdsp< N_SDSP; ++sdsp) {
			if (!(hc.slvBits & (1<<sdsp))) continue; 
			setSlvReg(sdsp, TRAP_CMD_STAT, (0x0 | (1<<TCSR_ISR_PENDING)) );
		}
	
		initTime= TIMER_getCount(timer1);
		do {
			errorCode[3]= chkSlaves(hc.slvBits, &hc.ready, HSTATUS_REG_0, HSR0_SLV_RDY);
		} while (  (delta_t(initTime) < HCTRL_TIMEOUT)
		         &&(hc.slvBits != hc.ready) );
	
		if (FATAL(errorCode[3])) {
			addError(&returnCode, errorCode[3], "histoCtrlInit", "chkSlaves",
			         __FILE__, __LINE__);
			return returnCode;
		}
		else if (hc.ready != hc.slvBits) {
			sprintf(genStr, "%s%s%d%s","The histogramming task is not running on ",
				        " one of the slaves, or it is not under master control.\n");
			newError(&returnCode, SLAVE_NOT_READY, FATAL_ERR, "histoCtrlInit",
			 genStr,  __FILE__, __LINE__);
			return returnCode;
		}
	}

	return returnCode;
}
#endif


/************************************************************************************
 * histoCtrlNewBin
 *   
 *   synopsis: Configures the modules in preparation for pulsing, and signals the
 *             slave DSPs to begin expecting events. After adjusting the scan
 *     variable, NewBin loads the entire module configuration structure (either
 *     basic, trim or both) into the modules. The routine then sends a soft-reset
 *     to the modules, and creates the trigger stream from the trigger sequence
 *     command lists.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlNewBin(void) {
	INT32  returnCode= SUCCESS, errorCode;
	UINT32 i, j, t0; //, val;
	UINT8  slv, k, bin, sendCfg;
	UINT8  chip, m;
	UINT32 mval;

	t0= TIMER_getCount(timer1);
	hc.timing->nb.nExe++;
	hc.timing->nb.dt_task+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	/* General setup: set the current command masks to the current run-time masks, and
	   update all the MDSP registers & counters. Then set up all the necessary registers on
	   each SDSP we are working with. The SDSPs regard each pass through NEWBIN as just
	   that-- a new bin (albeit sometimes the bin # stays the same). */
	setTaskState(HISTOGRAM_CTRL_TASK, HCTRL_NEWBIN);
	getTaskCtrlBits();

	if (hc.binPause) scanWait(SCAN_NB_INIT, 0);

	/* set initial command and data mask states */
	if (!GET_RBIT(DIAGNOSTIC_REG, DR_USE_MDSP_MASK_LUT))
		writeRegisterDirect(FE_MASK_LUT_SELECT, 3, 0, MASK_SET_INIT);
	else {
		switchLinkMasks(((DATA_LINK)+(COMMAND_LINK)), MASK_SET_INIT);

		if (  (hc.cfgReg[0] == ST_BYPASS) || (hc.cfgReg[0] == ST_TOKEN)
		    ||(hc.cfgReg[1] == ST_BYPASS) || (hc.cfgReg[1] == ST_TOKEN) ) {
			/* Pre-compute the mask-set deltas for each SDSP if the masks will be changing;
			note that we only care about the serial port command link masks and the dynamic
			mode bits (DATA_LINK). The link configuration (on or off) is left unchanged by
			these mask sets to increase switching speed. This is done in a separate SDSP
			loop to allow paired SDSPs to be linked together. */
			for (i= 0; i < N_SDSP; ++i) {
				setLinkMasks(0, SP_BOTH, ((DATA_LINK) +(COMMAND_LINK)), COMPUTE_MASK_DELTAS, i);
			}
	
			setLinkMasks(0, SP_BOTH, ((DATA_LINK) +(COMMAND_LINK)), COMPUTE_MASK_DELTAS,
			             MASK_SET_INIT);
		}
	}

	//Reset the EFB counters by toggling the cmd. reg. bit:
	writeRegister(EFB_CMND_0, 1, EFB_GROUP_COUNTER_EN_O, 0);
	writeRegister(EFB_CMND_0, 1, EFB_GROUP_COUNTER_EN_O, 1);
	hc.efbCnt= 0x00000000;
	hc.evtWait= TRUE;   hc.trigLoopMax= 1;   hc.trigDelay= 0;
	newTrapSeries(); //Give the SDSPs a new trap series.

	bin= hc.status->currentBin[0];
	ASGN_RFLD(HSTATUS_REG_0, HSTAT0_BIN, HSTAT0_BIN_W, bin);
	//ASGN_RFLD(HSTATUS_REG_0, HSTAT0_CHIP, HSTAT0_CHIP_W, hc.status->currentChip);
	ASGN_RFLD(HSTATUS_REG_0, HSTAT0_STAGE, HSTAT0_STAGE_W, hc.status->currentMaskStage);

	WRITE_REG(HSTATUS_REG_1, 0);
	WRITE_REG(HSSTAT_REG_0,     0);		WRITE_REG(HSSTAT_REG_1,     0);
	WRITE_REG(HSSTAT_REG_2,     0);		WRITE_REG(HSSTAT_REG_3,     0);

	/* reset all the counters: */
	hc.pulsing= hc.slvBits;
	for (slv= 0; slv< N_SDSP; ++slv) {
		if ((hc.scan->reset.binReset) && (hc.scan->reset.ECR)) {
			hc.dspInfo[slv].L1id[0]=    hc.dspInfo[slv].L1id[1]= 0; 
		}
		//hc.dspInfo[slv].nEvents[0]= hc.dspInfo[slv].nEvents[1]= 0;
		//hc.dspInfo[slv].nErrors[0]= hc.dspInfo[slv].nErrors[1]= 0;
		hc.dspInfo[slv].nTrig= hc.dspInfo[slv].nEvents= hc.dspInfo[slv].nErrors= 0;
		hc.dspInfo[slv].delErr= 0;
		hc.dspInfo[slv].updateCnt= hc.dspInfo[slv].sigmaEvtLen= 0;
		hc.dspInfo[slv].avgEvtLen= hc.dspInfo[slv].avgProcTime= 0;
	}

#if 0
	val=  (hc.status->currentBin[0]<<HCMD_SLV_BIN) +(1<<HCMD_SLV_NEWBIN)
		 +(hc.status->currentMaskStage<<HCMD_SLV_STAGE);
	if (hc.manual) scanWait(SCAN_SDSP_SETUP, 0);

	hc.timing->nb.dt_setup+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	/* set the slaves histogram command registers. The waitExp state waits until they
	   indicate they are ready to begin histogramming. */ 
	for (slv= 0; slv< N_SDSP; ++slv) {
		if (!(hc.slvBits & (1<<slv))) continue; 

		setSlvReg(slv, HCMD_REG, val);
	}


#else
	//dpsf temp:
	hc.calLineLoop= (hc.scan->general.maskMode == SCAN_STAGED);
	hc.calLine= hc.status->currentMaskStage;
	for (slv= 0; slv< N_SDSP; ++slv) {
		if (!(hc.slvBits & (1<<slv))) continue; 

		/* override the task manager's event filter on the slave temporarily */
		setSlvRegBit(slv, COMMAND_REG_0, CR_HISTO_OVERRIDE); //dpsf temp.

		setSlvReg(slv, HCMD_STAT_REG_1, (hc.slvRep<<HCSR1_SLV_BINEVT));

		if (hc.calLineLoop) {
			setSlvReg(slv, HCMD_STAT_REG_0,
			               ( (hc.status->currentBin[0]<<HCSR0_SLV_BIN)
			                +(1<<HCSR0_SLV_NEWBIN)
			                +(hc.calLineLoop<<HCSR0_SLV_CALLINE_EN)
			                +(hc.calLine<<HCSR0_SLV_CALLINE)
			                +(1<<HCSR0_SLV_NEWCAL)                 ) );
		}
		else {
			setSlvReg(slv, HCMD_STAT_REG_0,
			               ( (hc.status->currentBin[0]<<HCSR0_SLV_BIN)
			                +(1<<HCSR0_SLV_NEWBIN)  ) );
		}
	}
#endif


	//hc.timing->nb.dt_sdsp_setup+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);


#if 0
	/* If advancing the current mask stage, calculate the new parameters, and then
	   re-configure the masks on the modules.  dpsf: for pixels & SCT, use pre-calculated
	   global masks-- do not store the masks separately each time on each module-- time
	   intensive. Instead, the current scan mask stage indicates the current mask, and
	   it is broadcast to all modules, and then latched into the chips. If a given
	   module needs special treatment: SCT: its mask is taken out of the broadcast
	   and then each chips channel mask is set by bitwise ORing each channel's mask
	   with the global mask.   must decouple mask from basic struct sending in sendCfgSet.
	   PXL: The FE enable on each module is looked at, if a given module has some FE
	   not enabled, its mask is taken out of the global mask, the broadcast is done, with
	   no mask setting/restoring in sendCfgSet & new settings : broadcast flag: loop
	   disabled, will use the global structure's settings as reference, and add in the
	   latch for each enabled chip only. */
	if (hc.stageAdvance) {
		hc.timing->nb.loop[0].nExe++;

		//#ifndef SIM
		hc.timing->nb.loop[0].nExe++;
		if (hc.stagePause) scanWait(SCAN_STAGE_CFG, 0);

		for (i=0; i<N_MODULES; ++i) {
		if ((hc.moduleCfg[i].present)&&(hc.moduleCfg[i].active)) {
			errorCode= setChipVariable(&hc.moduleCfg[i], ALL_CHIPS, ST_CAL_MODE,
			                           (1.0*hc.status->currentMaskStage));

/*
			errorCode= setNextMaskStage(&hc.moduleCfg[i],
			                            hc.status->currentChip, 
			                            hc.scan->general.moduleScanMode,
			                            hc.status->currentMaskStage);
if not broadcasting
*/
		}}
		//hc.timing->nb.loop[0].dt_calc[0]+= delta_t(hc.time);

		if (hc.stagePause) scanWait(SCAN_STAGE_SEND, 0);
//Modify to allow sendCfg flag (set here) to handle stage setting while congfiguring
//if it is doing so (if not configuring, the stage is set here):
		errorCode= sendConfigSet(hc.port,  hc.mod8, ALL_CHIPS, //hc.status->currentChip, 
			                         NORMAL_CONFIG_LOOP, //FALSE /*broadcast*/,
			                         TRUE, /* set & restore links */
			                         hc.cfgSet, MODULE_GROUP_ALL, hc.stageCfg,
			                         TRUE /*activeOnly*/, FALSE /*enableDataTaking*/);

		//hc.timing->nb.loop[0].dt_calc[1]+= cfgTiming.setup;
		//hc.timing->nb.loop[0].dt_cfg+=     cfgTiming.cfg;
		//hc.timing->nb.loop[0].total+= delta_t(hc.time);
		//#endif
	}

#else
	for (i=0; i<N_MODULE_GROUPS; ++i) {
		for (j=0; j<N_MODULES; ++j) {
		if ((hc.moduleCfg[j].present) && (hc.moduleCfg[j].groupId == i)) {
			if (hc.calLineLoop) mval= 0x11111111<<hc.calLine;

			for (chip= 0; chip< N_CHIPS; ++chip) {
				if (hc.calLineLoop) {
					errorCode= setChipVariable(&hc.moduleCfg[j], chip,
					                            ST_CAL_MODE, (1.0*hc.calLine));
					for (m=0; m<4; ++m) {
						hc.moduleCfg[j].chip[chip].basic.mask[m]= mval;
					}
				}
			}
		
		}}  /* loop over modules (present & in group) */
	} /* loop over groups */

	if (hc.stageAdvance) {
		errorCode= sendConfigSet(hc.port,  hc.mod8, ALL_CHIPS, //hc.status->currentChip,
		                         NORMAL_CONFIG_LOOP, //FALSE /*broadcast*/,
		                         TRUE, /* set & restore links */
		                         hc.cfgSet, MODULE_GROUP_ALL, CONFIG_MODULE_BASIC,
		                         FALSE /*activeOnly*/, TRUE /*enableDataTaking*/);
	}

#endif


	//#if !defined(SIM)
	/* load the variable into the configuration set (and set the calibration line
	   if enabled). This is actually done only if it makes sense (time-wise) to do
	   so, for varying things like trim-DACs the routine actually only loads a global
	   structure. For varying the ROD trigger delay, the current delay in the scan
	   control structure is updated instead: */
	if (hc.manual) scanWait(SCAN_CFG_VARIABLE, 0);
	hc.time= TIMER_getCount(timer1);

	sendCfg= FALSE;
	for (i=0; i<2; ++i) {
	if (hc.binAdvance[i]) {
		if (  (hc.scan->general.scanParameter[i] == ST_TRIG_DELAY1)
	        ||(hc.scan->general.scanParameter[1] == ST_TRIG_DELAY2) ) {

			hc.status->currentDelay= varRange[i][hc.status->currentBin[i]];
			if (hc.scan->serial.calcFromStruct) buildScanTrigger(FALSE);
			else  modifyTriggerDelay(hc.scan->general.scanParameter[i]);
		}

		else {
			sendCfg= TRUE;

			for (j=0; j<N_MODULES; ++j) {
			if ((hc.moduleCfg[j].present) && (hc.moduleCfg[j].active)) {
				/* If a 2-D scan is being performed, set the modules' loop variables (if
				   needed) to the new values. If a 1-D scan is being done, the i = 1 set
				   of loop variables is used to represent an alternate set of points for
				   the loop variable. This is typically used for current load balancing
				   for the SCT. */
	
				if (hc.twoDScan) k= i;
				else             k= hc.groupInfo[hc.moduleCfg[j].groupId].rangeList;

//dpsf: currentChip must be there to accomodate token & bypass?	
#if 0
				errorCode= setChipVariable(&hc.moduleCfg[j], hc.status->currentChip,
				                           hc.cfgReg[i], 
				                           varRange[k][hc.status->currentBin[i]]);
#else
//dpsf: temp.
				for (chip= 0; chip< N_CHIPS; ++chip) {
					/* If we are scanning across variables such as ST_BYPASS, TOKEN or 
					   a BOC variable, no need to loop over the chip variable. */
					if ((chip>0) && (hc.singleCfgVar[i])) continue;

					errorCode= setChipVariable(&hc.moduleCfg[j], chip,
					                           hc.cfgReg[i], 
					                           varRange[k][hc.status->currentBin[i]]);
				}

#endif

			}} // Loop over modules (present & in group)
		}

		//hc.timing->nb.loop[1+i].dt_calc[0]+= delta_t(hc.time);
	}} // Loop over bin indices which advance.
	//#endif

	if (hc.manual) scanWait(SCAN_SEND_CFG, 0);
	/* later: if this RODs EXTERNAL MODULE histogram cmd reg flag has been set,
	   then the ROD will pause between bins & signal the SBC that its ext modules
	   need re-configuring by the controlling ROD(s?!!). All RODS with flag set
	   will increment bins in lockstep, and an additional sendConfigSet call is
	   made here. Histogram event triggering for these RODs is done by the SBC/TIM.
	   If an error event occurs during histogramming, the ROD will notify the SBC
	   which will send appropriate commands to the controlling ROD(s) (soft reset,
	   etc.) */

	if (sendCfg) {
		errorCode= sendConfigSet(hc.port,  hc.mod8, ALL_CHIPS, //hc.status->currentChip,
		                         NORMAL_CONFIG_LOOP, //FALSE /*broadcast*/,
		                         TRUE, /* set & restore links */
		                         hc.cfgSet, MODULE_GROUP_ALL,     hc.dataType[0],
		                         FALSE /*activeOnly*/, TRUE /*enableDataTaking*/);
	}
	
	hc.timing->nb.dt_calc+= cfgTiming.setup;
	hc.timing->nb.dt_cfg+=  cfgTiming.cfg;

	for (i=0; i<2; ++i) {
	if (hc.binAdvance[i]) {
		//hc.timing->nb.loop[1+i].dt_calc[1]+= cfgTiming.setup;
		//hc.timing->nb.loop[1+i].dt_cfg+=     cfgTiming.cfg;
		//hc.timing->nb.loop[1+i].total+= delta_t(hc.time);
	}}

	hc.time= TIMER_getCount(timer1);
	
	if (hc.manual) scanWait(SCAN_NB_ECR_BCR, 0);
	if (  (hc.scan->reset.binReset)
	    &&((hc.scan->reset.ECR) || (hc.scan->reset.BCR)) ){
		resetModules(0, hc.scan->reset.ECR, hc.scan->reset.BCR, TRUE, TRUE);
	}

	hc.timing->nb.dt_resets+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	hc.first= TRUE;

	/* Set the last trigger time for all SDSPs to the current time 
	   + i x the dwelling time per dsp: */
	for (i=0; i<N_SDSP; ++i)
		//hc.timing->pulse[bin].dsp[i].lastTrigger=
		//	hc.time +i*(hc.timing->pulse[bin].dspDwell);
		;

	//Add in total time taken in the routine:
	hc.timing->nb.total+= delta_t(t0);
	hc.time= TIMER_getCount(timer1);

	return returnCode;
}

#define WAITING_FOR_SLAVES   1
/************************************************************************************
 * histoCtrlWaitExp
 *   
 *   synopsis: Waits until the slave DSPs indicate that they are expecting events.
 *             Normally this state will exit immediately since the slaves' registers
 *     were set in the beginning of state NEWBIN, and several micro-seconds have
 *     elsapsed. However, to account for the possibility that the slaves have some
 *     initializations to do before they are ready, and to allow any interleaved
 *     primitives to execute, a pause for up to one second is allowed; the master DSP
 *     will repeat this state until the time limit is reached, or the slaves are
 *     expecting events. 
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlWaitExp(void) {
	INT32  returnCode= WAITING_FOR_SLAVES, status;
	static UINT32 t0;
	UINT8  slv;

	if (hc.first) {
		hc.first= FALSE;
		setTaskState(HISTOGRAM_CTRL_TASK, HCTRL_WAITEXP);
		t0= TIMER_getCount(timer1);

		hc.timing->wexp.nExe++;
		hc.timing->wexp.dt_task+= delta_t(hc.time);
		hc.time= TIMER_getCount(timer1);
	}

	status= chkSlaves(hc.slvBits, &hc.expecting, SDSP_HSTATUS_REG_0, HSR0_SLV_EXP);
	if (status != SUCCESS) {
		addError(&returnCode, status, "histoCtrlWaitExp", "chkSlaves",
		         __FILE__, __LINE__);
		if (FATAL(returnCode)) return returnCode;
	}
	
	if (hc.expecting == hc.slvBits) {
		hc.first= TRUE;
		returnCode= SUCCESS;

		/* reset the slaves' NEWBIN & NEWCAL bits */
		for (slv= 0; slv< N_SDSP; ++slv) {
			if (!(hc.slvBits & (1<<slv))) continue; 
			rstSlvRegBit(slv, HCMD_STAT_REG_0, HCMD_SLV_NEWBIN);

#if 1
//dpsf: temp:
			if (hc.calLineLoop) {
				rstSlvRegBit(slv, HCMD_STAT_REG_0, HCSR0_SLV_NEWCAL);
			}
			/* turn off override: the slaves will only enter task when an event needs
			   processing */
			rstSlvRegBit(slv, COMMAND_REG_0, CR_HISTO_OVERRIDE);
#endif
		}

		//Add in total time taken in the routine:
		hc.timing->wexp.total+= delta_t(t0);
		hc.time= TIMER_getCount(timer1);
	}

	else if (delta_t(t0) > HCTRL_TIMEOUT) {
		if (GET_RBIT(DIAGNOSTIC_REG, DR_HISTO_3)) scanWait(SCAN_WAITEXP, 0);

		sprintf(genStr, "%s%s%s%s0x%04x, 0x%04x%s","A problem has developed on ",
			"(at least) one of the slaves or it is not under master control.\n",
			"Timed out while waiting for slave to indicate it was ready.\n",
			"Slave, expecting bitfields= ",hc.slvBits, hc.expecting,".\n");
		newError(&returnCode, SLAVE_NOT_READY, FATAL_ERR, "histoCtrlWaitExp",
		          genStr,  __FILE__, __LINE__);
	}

	return returnCode;
}

/************************************************************************************
 * histoCtrlWaiting
 *   
 *   synopsis: Waits until the slave DSPs indicate that they have completed
 *             processing this bin. Normally this state will exit immediately since
 *     the slaves have finished with their events. However, to account for the
 *     possibility that the slaves have cleanup to do after the bin has ended, and
 *     to allow any interleaved primitives to execute, a pause for up to ten seconds
 *     is allowed; the master DSP will repeat this state until the time limit is
 *     reached, or the slaves indicate that they are done. 
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlWaiting(void) {
	INT32  returnCode= WAITING_FOR_SLAVES, status;
	static UINT32 t0;

	if (hc.first) {
		hc.first= FALSE;
		setTaskState(HISTOGRAM_CTRL_TASK, HCTRL_WAITING);
		t0= TIMER_getCount(timer1);

		hc.timing->wait.nExe++;
		hc.timing->wait.dt_task+= delta_t(hc.time);
		hc.time= TIMER_getCount(timer1);
	}

	status= chkSlaves(hc.slvBits, &hc.done, SDSP_HSTATUS_REG_0, HSR0_SLV_DONE);
	if (status != SUCCESS) {
		addError(&returnCode, status, "histoCtrlWaiting", "chkSlaves",
		         __FILE__, __LINE__);
		if (FATAL(returnCode)) return returnCode;
	}

	if (hc.done == hc.slvBits) {
		hc.first= TRUE;
		returnCode= SUCCESS;

		hc.timing->wait.total+= delta_t(t0);
		hc.time= TIMER_getCount(timer1);
	}

	else if (delta_t(t0) > HCTRL_TIMEOUT) {
		sprintf(genStr, "%s%s%s%s0x%04x, 0x%04x%s","A problem has developed on ",
			"(at least) one of the slaves or it is not under master control.\n",
			"Timed out while waiting for slave to indicate it was done.\n"
			"Slave, done bitfields= ",hc.slvBits, hc.done,".\n");
		newError(&returnCode, SLAVE_NOT_READY, FATAL_ERR, "histoCtrlWaiting",
		 genStr,  __FILE__, __LINE__);
	}

	return returnCode;
}

/************************************************************************************
 * histoCtrlPrep
 *   
 *   synopsis: Prepares the MDSP to enter NEWBIN. On the MDSP (& SDSPs), the NEWBIN
 *             state simply implies that a variable has changed-- the actual bin
 *     number may or may not change. The routine increments the current mask stage or
 *     bin number, and perhaps both. NEWBIN does all the actual mask setting on the
 *     modules. Once all bins are completed, the routine flags the task to exit.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
UINT8 histoCtrlPrep(void) {
	UINT8  state, i, j, done= FALSE;

	/* The necessary adjustments to the current mask stage, inner and outer loop
	   variables, and chip are made here & tested to see if the scan has finished.
	   If not done, NEWBIN will transfer the new settings (i.e. stage, etc.) to
	   the modules. If mask staging is requested, masks can be staged either before
	   or after the inner loop variable, depending on the boolean stageAdvanceFirst 
	   variable's setting. */
	setTaskState(HISTOGRAM_CTRL_TASK, HCTRL_PREP);

	hc.timing->prep.nExe++;
	hc.timing->prep.dt_task+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);

	hc.stageAdvance= hc.binAdvance[0]= hc.binAdvance[1]= FALSE;

	if (hc.scan->general.maskMode == SCAN_STAGED) {
		if (hc.scan->general.stageAdvanceFirst) { //Advance stage 1st
			hc.stageAdvance= TRUE;
			if (hc.status->currentMaskStage >= (hc.scan->general.maskStages -1))
				hc.binAdvance[0]= TRUE;

			if (  (hc.binAdvance[0])
			    &&(hc.status->currentBin[0] >= (hc.scan->general.nBins[0] -1)) )
				hc.binAdvance[1]= TRUE;
		}

		else {  //Advance bin[0] 1st
			hc.binAdvance[0]= TRUE;
			if (hc.status->currentBin[0] >= (hc.scan->general.nBins[0] -1))
				hc.stageAdvance= TRUE;

			if (  (hc.stageAdvance)
			    &&(hc.status->currentMaskStage >= (hc.scan->general.maskStages -1))
			   )
				hc.binAdvance[1]= TRUE;
		}
	}

	else {
		hc.binAdvance[0]= TRUE;
		if (hc.status->currentBin[0] >= (hc.scan->general.nBins[0] -1))
			hc.binAdvance[1]= TRUE;
	}

	/* Get new settings */
	if (hc.stageAdvance) {
		++hc.status->currentMaskStage;
		if (hc.status->currentMaskStage >= hc.scan->general.maskStages)
			hc.status->currentMaskStage= 0;
	}

	for (i=0; i<2; ++i) {
		if (hc.binAdvance[i]) {
			++hc.status->currentBin[i];
			if (hc.status->currentBin[i] >= hc.scan->general.nBins[i]) {
				hc.status->currentBin[i]= 0;
				if (i == 1) done= TRUE;
			}
		}
	}


	if (!done) state= HCTRL_NEWBIN;
	else       state= TASK_DONE;

	hc.timing->prep.total+= delta_t(hc.time);
	hc.time= TIMER_getCount(timer1);
	return state;
}

/************************************************************************************
 * histoCtrlDone
 *   
 *   synopsis: Loads the output structure variables, and halts histogramming and
 *             data-taking on the slaves if they were not externally set up. The ROD
 *     is returned to the state in which it was when the task was initially entered
 *     (link setup, FPGA registers, etc.) This routine halts the scan gracefully.
 *     There are several ways in which this routine can be reached:
 *
 *          1) via sucessful completion of the task.
 *          2) a fatal run-time error.
 *          3) user abort of a scan.
 *
 *     In case 1 & 2 the routine will actually be called twice: once when the
 *     DONE state is entered, and once again when the task manager senses that
 *     the task has requested termination and sets the opFlag to TASK_HALT. In
 *     case 3, the latter call above is the only call made. In the initial call
 *     the routine uses the error flag in the histogram control structure to
 *     determine the overall error status of the task, and returns the ROD state
 *     to the state it was in at the time the task was initialized. Any subsequent
 *     calls (before the task has been re-initialized) will be treated as dummy
 *     calls.
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlDone(void) {
	INT32 status;

	if (!hc.reinit) {
		hc.reinit= TRUE;

		hc.timing->done.dt_task+= delta_t(hc.time);
		hc.time= TIMER_getCount(timer1);
	
		/* Restore the original masks & running mode. */
		setLinkMasks(0, SP_BOTH, ((DATA_LINK)+(COMMAND_LINK)), SET_MASK, MASK_SET_INIT);

		hc.timing->done.total+= delta_t(hc.time);
		hc.time= TIMER_getCount(timer1);
		return hc.error;
	}
	else {
		return SUCCESS;
	}
}

typedef struct {
	UINT32 issueTrig, trigLoopMax, trigSent,
	       efbCnt, cnt, icnt1, icnt2,
	       sdspMask, mask, irdy, ready;
} xStruct;
xStruct x;

/************************************************************************************
 * histoCtrlTask
 *   
 *   synopsis: Controls & monitors the filling of slave DSP histograms during a scan.
 *             The task (state-machine) states which are infrequently called are
 *     handled by subroutines (so they can be put in different memory segments if
 *     desired); the pulsing state is the only state which is not done via a separate
 *     routine as the additional delay to set up & call it (up to 2-3 micro-seconds)
 *     is potentially undesirable. The smaller routine is also easier to fit in the
 *     MDSP's fast internal memory.
 *
 *         The routine is centered around the concept of module groups: sets of modules
 *     which get the same trigger & behave in a similar fashion during a scan. Which
 *     group a given module is in is determined by that module's (pre-loaded) control
 *     structure (eg. ABCDModule).
 *
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 histoCtrlTask(union TASK_STRUCTURES_IN *tsi, union TASK_STRUCTURES_OUT *tso, UINT32 opFlags) {
	INT32  returnCode= REPEAT_TASK,  status, errorCnt;
	UINT32 t0, dt;
	static UINT8  state= TASK_INIT;

	register UINT32 sdsp, twin;
#if defined(XS)
	register UINT32 issueTrig, trigLoopMax, trigSent,
	                efbCnt, cnt, icnt1, icnt2,
	                sdspMask, mask, irdy, ready;
#endif

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {return (returnCode= scanTaskOp(tso, opFlags, &state)); }

	/************************* routine states:PULSING *************************/
	if (state == HCTRL_TRIGGERS) {
		t0= TIMER_getCount(timer1);

		/* The trigSent variable is a bitfield based upon the corresponding
		   number of evts sent, with 8 bits / sdsp. It is compared inside the
		   loop to the trigLoopMax variable. */ 
#if defined(XS)
		trigSent= 0x00000000;
		trigLoopMax= hc.trigLoopMax;    //controls # triggers done this iteration.
		efbCnt=      hc.efbCnt;         //retrieve stored event count.
		issueTrig=   hc.pulsing;        //bitfield controlling trigger sending.
#else
		x.trigSent= 0x00000000;
		x.trigLoopMax= hc.trigLoopMax;    //controls # triggers done this iteration.
		x.efbCnt=      hc.efbCnt;         //retrieve stored event count.
		x.issueTrig=   hc.pulsing;        //bitfield controlling trigger sending.
#endif

#if defined(XS)
		while (issueTrig) {
#else
		while (x.issueTrig) {
#endif
			//Adjust trigger timing if needed:
			if (hc.trigDelay) delay(hc.trigDelay);

			//Cannot use pointer to register variable, so borrow dt:
			status= readRegister(EFB_EVT_CNT, 32, 0, &dt);
#if defined(XS)
			cnt= dt;

			//Use extu intrinsic to guarantee single cycle for extract:
			icnt1= _extu(cnt, 28, 28);	icnt2= _extu(efbCnt, 28, 28);
			ready= (icnt1 == icnt2);								//sdsp 0

			icnt1= _extu(cnt, 24, 28);	icnt2= _extu(efbCnt, 24, 28);
			irdy= (icnt1 == icnt2);		irdy<<= 1;	ready|= irdy;	//sdsp 1

			icnt1= _extu(cnt, 20, 28);	icnt2= _extu(efbCnt, 20, 28);
			irdy= (icnt1 == icnt2);		irdy<<= 2;	ready|= irdy;	//sdsp 2

			icnt1= _extu(cnt, 16, 28);	icnt2= _extu(efbCnt, 16, 28);
			irdy= (icnt1 == icnt2);		irdy<<= 3;	ready|= irdy;	//sdsp 3
#else
			x.cnt= dt;

			//Use extu intrinsic to guarantee single cycle for extract:
			x.icnt1= _extu(x.cnt, 28, 28);	x.icnt2= _extu(x.efbCnt, 28, 28);
			x.ready= (x.icnt1 == x.icnt2);								//sdsp 0

			x.icnt1= _extu(x.cnt, 24, 28);	x.icnt2= _extu(x.efbCnt, 24, 28);
			x.irdy= (x.icnt1 == x.icnt2);		x.irdy<<= 1;	x.ready|= x.irdy;	//sdsp 1

			x.icnt1= _extu(x.cnt, 20, 28);	x.icnt2= _extu(x.efbCnt, 20, 28);
			x.irdy= (x.icnt1 == x.icnt2);		x.irdy<<= 2;	x.ready|= x.irdy;	//sdsp 2

			x.icnt1= _extu(x.cnt, 16, 28);	x.icnt2= _extu(x.efbCnt, 16, 28);
			x.irdy= (x.icnt1 == x.icnt2);		x.irdy<<= 3;	x.ready|= x.irdy;	//sdsp 3
#endif

			/* If any SDSP is indicating that it's not ready, break out of loop. This
			   should only happen with a) an error or b) the SDSPs are using slower
			   histogramming code & are filling up. Note that all the SDSPs are checked,
			   even if some are not participating in the scan, since those SDSP could be
			   set up externally to receive scan events as well. The condition is handled
			   in the scanAttnHandle routine. */
			status= readRegister(INTRPT_FROM_SLV, 4, 0, &hc.sdspAttention);
			if (hc.sdspAttention != 0xf) break;

			hc.trigTime= TIMER_getCount(timer1);
#if defined(XS)
			ready &= issueTrig;
#else
			x.ready &= x.issueTrig;
#endif				

			for (sdsp= 0; sdsp<N_SDSP; ++sdsp) {
				/* The mask bitfield has 1<<(sdsp) set for any active sdsp (& another
				   bit set for the twinned sdsp if paired); for paired sdsps which
				   are passively receiving triggers, the mask is set to an unphysical
				   value so that the following inequality will always return true: */
#if defined(XS)
				mask= hc.dspInfo[sdsp].mask;
				if ((ready & mask) != mask) continue;
#else
				x.mask= hc.dspInfo[sdsp].mask;
				if ((x.ready & x.mask) != x.mask) continue;
#endif
				if (hc.dspInfo[sdsp].paired) twin= hc.dspInfo[sdsp].pair;
				else                         twin= sdsp;


				/* Two Reasons to reset the MDSP's L1 ID counter:
				   1) if a ECR (soft reset) has been added into the trigger stream
				      before each event or
				   2) if the DSP is using simulated events which were pre-loaded in
				      the inmems (hence a fixed L1 ID (the BC ID will also be fixed). 
				   These are built into the L1idReset flag. */
				if (hc.L1idReset) {hc.dspInfo[sdsp].L1id[0]= hc.dspInfo[sdsp].L1id[1]= 0; }
				writeRegisterDirect(CAL_L1_ID_0, 32, 0, hc.dspInfo[sdsp].L1id[0]);
				writeRegisterDirect(CAL_L1_ID_1, 32, 0, hc.dspInfo[sdsp].L1id[1]);

				writeRegisterDirect(DFLT_ROD_EVT_TYPE, 32, 0, sdsp);
				writeRegisterDirect(CRTV_ROD_EVT_TYPE, 32, 0, twin);

				if (!GET_RBIT(DIAGNOSTIC_REG, DR_USE_MDSP_MASK_LUT))
					writeRegisterDirect(FE_MASK_LUT_SELECT, 3, 0, sdsp);
				else
					switchLinkMasks(((DATA_LINK)+(COMMAND_LINK)), sdsp);

				//Send a trigger to the modules; the event will go to the current sdsp:
				simTriggerArmM(sdsp);
				serialStreamOut(0, (TransData *) hc.ptr, hc.nSets);
				
				++hc.dspInfo[sdsp].L1id[0];
				++hc.dspInfo[sdsp].L1id[1];

#if defined(XS)
				icnt1= sdsp<<3;   icnt2= 24 -icnt1;
				trigSent+= 1<<icnt1;
				cnt= _extu(trigSent, icnt2, 24);
				if (cnt == trigLoopMax) issueTrig= _clr(issueTrig, sdsp, sdsp);

				//Update the EFB counter (4 bits => must handle wrapping):
				icnt1= sdsp<<2;   icnt2= icnt1 +3;  //borrow icnt1, icnt2 & cnt.
				//mask= 0xf<<icnt1;  mask= ~mask;
				cnt= 28 -icnt1;
				cnt=    _extu(efbCnt, cnt, 28);
				efbCnt= _clr(efbCnt, icnt1, icnt2);
				++cnt;   cnt&= 0xf;   cnt<<= icnt1;
				efbCnt|= cnt;  

				if (hc.dspInfo[sdsp].paired) {
					trigSent+= 1<<(twin<<3);

					icnt1= twin<<2;   icnt2= icnt1 +3;
					cnt= 28 -icnt1;
					cnt=    _extu(efbCnt, cnt, 28);
					efbCnt= _clr(efbCnt, icnt1, icnt2);
					++cnt;   cnt&= 0xf;   cnt<<= icnt1;
					efbCnt|= cnt;  
				}
#else
				x.icnt1= sdsp<<3;   x.icnt2= 24 -x.icnt1;
				x.trigSent+= 1<<x.icnt1;
				x.cnt= _extu(x.trigSent, x.icnt2, 24);
				if (x.cnt == x.trigLoopMax) x.issueTrig= _clr(x.issueTrig, sdsp, sdsp);

				//Update the EFB counter (4 bits => must handle wrapping):
				x.icnt1= sdsp<<2;   x.icnt2= x.icnt1 +3;  //borrow icnt1, icnt2 & cnt.
				//mask= 0xf<<icnt1;  mask= ~mask;
				x.cnt= 28 -x.icnt1;
				x.cnt=    _extu(x.efbCnt, x.cnt, 28);
				x.efbCnt= _clr(x.efbCnt, x.icnt1, x.icnt2);
				++x.cnt;   x.cnt&= 0xf;   x.cnt<<= x.icnt1;
				x.efbCnt|= x.cnt;  

				if (hc.dspInfo[sdsp].paired) {
					x.trigSent+= 1<<(twin<<3);

					x.icnt1= twin<<2;   x.icnt2= x.icnt1 +3;
					x.cnt= 28 -x.icnt1;
					x.cnt=    _extu(x.efbCnt, x.cnt, 28);
					x.efbCnt= _clr(x.efbCnt, x.icnt1, x.icnt2);
					++x.cnt;   x.cnt&= 0xf;   x.cnt<<= x.icnt1;
					x.efbCnt|= x.cnt;  
				}
#endif

			} //sdsp loop

			//exit if time limit exceeded; this is typically set to .5 sec.
			dt= delta_t(t0);
			if (dt > hc.taskTimeMax) x.issueTrig= FALSE;
		} //send triggers?

#if defined(XS)
		hc.efbCnt= efbCnt;           //Store new EFB counter.

		/* Increment trigger counter; this is done per SDSP to allow flexibility
		   in deciding how to end the while loop above: */
		for (sdsp= 0; sdsp<N_SDSP; ++sdsp) {
			if (!(hc.slvBits & (1<<sdsp))) continue;
			icnt1= sdsp<<3;   icnt2= 24 -icnt1;
			hc.dspInfo[sdsp].nTrig+= _extu(trigSent, icnt2, 24);
		}

#else
		hc.efbCnt= x.efbCnt;           //Store new EFB counter.

		/* Increment trigger counter; this is done per SDSP to allow flexibility
		   in deciding how to end the while loop above: */
		for (sdsp= 0; sdsp<N_SDSP; ++sdsp) {
			if (!(hc.slvBits & (1<<sdsp))) continue;
			x.icnt1= sdsp<<3;   x.icnt2= 24 -x.icnt1;
			hc.dspInfo[sdsp].nTrig+= _extu(x.trigSent, x.icnt2, 24);
		}
#endif

		/* If requested (by scanExtrapolate), wait for events to fully clear the
		   event path. Watch until the efb bandwidth counter stops incrementing,
		   and then measure the total dt from the last event series to get an approx.
		   value for the event transmission time. Add on this value divided by 32
		   to allow for transmission into the router & SDSPs.
		   
		   For an SCT calibration, a typical 1/4 occupancy event in a low-threshold
		   bin will take ~17 bits per hit strip * 32 strips/chip * 6 chips/link
		   +34 (header + trailer) ~ 3300 bits, or 82.5 micro-seconds to transmit at
		   40 MHz. In the router & DSPs, the events should occupy about 3/4 of a frame
		   per module. One a link, it takes .85 micro-seconds to transmit one word.
		   Assuming 12 modules/24 links per SDSP, this translates into 28.2 words per
		   micro-second. */
#if defined(XS)
		if (hc.evtWait) {
			t0= TIMER_getCount(timer1);
			status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt1= dt;
			do {
				delay(20); //2 us delay.
				status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   icnt2= dt;
				ready= (icnt2 == icnt1); 
				icnt1= icnt2;
			} while ((!ready) && ((dt= delta_t(t0)) < HCTRL_TIMEOUT));
			cnt= delta_t(hc.trigTime);  icnt2= cnt>>4;
			delay(icnt2);
		}
#else
		if (hc.evtWait) {
			t0= TIMER_getCount(timer1);
			status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   x.icnt1= dt;
			do {
				delay(20); //2 us delay.
				status= readRegister(EFB_BANDWIDTH_CNT, 16, 0, &dt);   x.icnt2= dt;
				x.ready= (x.icnt2 == x.icnt1); 
				x.icnt1= x.icnt2;
			} while ((!x.ready) && ((dt= delta_t(t0)) < HCTRL_TIMEOUT));
			x.cnt= delta_t(hc.trigTime);  x.icnt2= x.cnt>>4;
			delay(x.icnt2);
		}
#endif

		//Now check the events & handle any errors:
		errorCnt= scanEventCheck(hc.evtWait);
		if (errorCnt) status= scanErrHandle();
		if (status != SUCCESS) {state= TASK_DONE; hc.error= status; }

		/* Check the attention flag & handle if necessary. Note that the raw
		   attention flag from the trigger loop is inverted: 0xf => all ok: */
		if (hc.sdspAttention != 0xf) status= scanAttnHandle();
		if (status != SUCCESS) {state= TASK_DONE; hc.error= status; }

		//Update statistics, and (potentially) extend the event loop's duration:
		updateScanStatistics();
		scanExtrapolate();

		if (!hc.pulsing) {/* yippie! */ hc.first= TRUE; state= HCTRL_WAITING; }
	} //state == trigger loop
	
	/*************************** other routine states *************************/
	/* Other routine states are handled in a separate routine to conserve
	   space in the on-chip program memory: */
	else {returnCode= scanTaskStates(tsi, &state); }

	return returnCode;
}

/************************************************************************************
 * scanTaskOp:      Handles output for the histogram control task.
 * scanTaskStates:  Handles all states other than the trigger loop for the
 *                  histogram control task
 ************************************************************************************/
INT32 scanTaskOp(union TASK_STRUCTURES_OUT *tso, UINT32 opFlags, UINT8 *state) {
	INT32 returnCode;
	tso->genTaskOut.dataPtr=     hc.status;
	tso->genTaskOut.dataLength=  SIZEOF(ScanStatus);

	if (opFlags & TASK_HALT_MASK) {
		newInformation(__FILE__, __LINE__, "Scan Control task halting.\n");
		*state= TASK_INIT;   returnCode= TASK_HALTED;
	}
	else if (opFlags & TASK_INIT_MASK) {
		newInformation(__FILE__, __LINE__, "Scan Control task re-initing.\n");
		*state= TASK_INIT;   returnCode= TASK_INITIALIZED;
	}
	else if (opFlags & TASK_QUERY_MASK) {
		newInformation(__FILE__, __LINE__, "Scan Control task: query data output.\n");
		returnCode= TASK_QUERIED;
	}
	else returnCode= REPEAT_TASK;
	
	return returnCode;
}

/************************************************************************************/
INT32 scanTaskStates(union TASK_STRUCTURES_IN *tsi, UINT8 *state) {
	INT32 returnCode= REPEAT_TASK, status;

	/*************************** routine states:INIT **************************/
	if (*state == TASK_INIT) {
		/* set up any static variables beyond state: */
		status= histoCtrlInit(tsi);
		if (status != SUCCESS) {*state= TASK_DONE; hc.error= status; }
		else *state= HCTRL_SDSP_SETUP;
	}

	/*************************** routine states:SDSP_SETUP ********************/
	 if (*state == HCTRL_SDSP_SETUP) {
		status= histoCtrlSdspSetup(tsi);
		if (status != SUCCESS) {*state= TASK_DONE; hc.error= status; }
		else *state= HCTRL_NEWBIN;
	}

	/************************** routine states:NEWBIN *************************/
	else if (*state == HCTRL_NEWBIN) {
		status= histoCtrlNewBin();
		if (status != SUCCESS) {*state= TASK_DONE; hc.error= status; }
		else *state= HCTRL_WAITEXP;
	}

	/************************* routine states:WAITEXP *************************/
	else if (*state == HCTRL_WAITEXP) {
		status= histoCtrlWaitExp();
		if (status != SUCCESS) {
			if (FATAL(status)) {
				*state= TASK_DONE; hc.error= status; return returnCode;
			}
			else return returnCode;  /* still waiting */
		}

		/* The state is set here to avoid having to either set it or test to
		   see if we've done so on each pass through the event loop: */
		*state= HCTRL_TRIGGERS;
		setTaskState(HISTOGRAM_CTRL_TASK, *state);
	}

	/************************* routine states:WAITING *************************/
	else if (*state == HCTRL_WAITING) {
		status= histoCtrlWaiting();
		if (status != SUCCESS) {
			if (FATAL(status)) {
				*state= TASK_DONE; hc.error= status; return returnCode;
			}
			else return returnCode;  /* still waiting */
		}
		*state= HCTRL_PREP;
	}

	/*************************** routine states:PREP **************************/
	else if (*state == HCTRL_PREP) {
		*state= histoCtrlPrep();    /* branch: DONE or NEWBIN */
		return returnCode;
	}

	/*************************** routine states:DONE **************************/
	else if (*state == TASK_DONE) {
		status= histoCtrlDone();
		returnCode= status;
	}

	return returnCode;
}
